from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Sequence, TypeVar

import numpy as np
from cmap import Colormap
from numpy.typing import ArrayLike, NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend, _text_utils
from whitecanvas.layers._base import HoverableDataBoundLayer
from whitecanvas.layers._mixin import (
    EdgeNamespace,
    FaceEdgeMixinEvents,
    FaceNamespace,
    MultiFaceEdgeMixin,
)
from whitecanvas.layers._primitive.text import Texts
from whitecanvas.layers._sizehint import xy_size_hint
from whitecanvas.protocols import MarkersProtocol
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    OrientationLike,
    Symbol,
    XYData,
    _Void,
)
from whitecanvas.utils.normalize import as_array_1d, normalize_xy, parse_texts

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers import group as _lg
    from whitecanvas.layers._mixin import ConstEdge, ConstFace, MultiEdge, MultiFace

_void = _Void()
_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)
_Size = TypeVar("_Size", float, NDArray[np.floating])


class MarkersLayerEvents(FaceEdgeMixinEvents):
    clicked = Signal(list)
    symbol = Signal(Symbol)
    size = Signal(float)


class Markers(
    HoverableDataBoundLayer[MarkersProtocol, XYData],
    MultiFaceEdgeMixin[_Face, _Edge],
    Generic[_Face, _Edge, _Size],
):
    events: MarkersLayerEvents
    _events_class = MarkersLayerEvents

    if TYPE_CHECKING:

        def __new__(cls, *args, **kwargs) -> Markers[ConstFace, ConstEdge, float]: ...

    def __init__(
        self,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        *,
        name: str | None = None,
        symbol: Symbol | str = Symbol.CIRCLE,
        size: float = 12.0,
        color: ColorType = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: Backend | str | None = None,
    ):
        MultiFaceEdgeMixin.__init__(self)
        xdata, ydata = normalize_xy(xdata, ydata)
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), xdata, ydata)
        self._size_is_array = False
        self.update(symbol=symbol, size=size, color=color, hatch=hatch, alpha=alpha)
        self.edge.color = color
        if not self.symbol.has_face():
            self.edge.update(width=2.0, color=color)
        pad_r = size / 400
        self._x_hint, self._y_hint = xy_size_hint(xdata, ydata, pad_r, pad_r)

        self._backend._plt_connect_pick_event(self.events.clicked.emit)
        self._init_events()

    @property
    def ndata(self) -> int:
        """Number of data points."""
        return self.data.x.size

    def _get_layer_data(self) -> XYData:
        return XYData(*self._backend._plt_get_data())

    def _norm_layer_data(self, data: Any) -> XYData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 2:
                raise ValueError("Expected a (N, 2) array")
            xdata, ydata = data[:, 0], data[:, 1]
        else:
            xdata, ydata = data
        x0, y0 = self.data
        if xdata is not None:
            x0 = as_array_1d(xdata)
        if ydata is not None:
            y0 = as_array_1d(ydata)
        if x0.size != y0.size:
            raise ValueError(
                "Expected xdata and ydata to have the same size, "
                f"got {x0.size} and {y0.size}"
            )
        return XYData(x0, y0)

    def _set_layer_data(self, data: XYData):
        x0, y0 = data
        self._backend._plt_set_data(x0, y0)
        if self._size_is_array:
            pad_r = self.size.mean() / 400
        else:
            pad_r = self.size / 400
        self._x_hint, self._y_hint = xy_size_hint(x0, y0, pad_r, pad_r)

    def set_data(
        self,
        xdata: ArrayLike | None = None,
        ydata: ArrayLike | None = None,
    ):
        if xdata is None:
            xdata = self.data.x
        if ydata is None:
            ydata = self.data.y
        self.data = XYData(xdata, ydata)

    @property
    def symbol(self) -> Symbol:
        """Symbol used to mark the data points."""
        return self._backend._plt_get_symbol()

    @symbol.setter
    def symbol(self, symbol: str | Symbol):
        sym = Symbol(symbol)
        if self.ndata > 0:
            self._backend._plt_set_symbol(sym)
        self.events.symbol.emit(sym)

    @property
    def size(self) -> _Size:
        """Size of the symbol."""
        size = self._backend._plt_get_symbol_size()
        if self._size_is_array:
            return size
        elif size.size > 0:
            return size[0]
        else:
            return 10.0  # default size

    @size.setter
    def size(self, size: _Size):
        """Set marker size"""
        ndata = self.ndata
        if not isinstance(size, (float, int, np.number)):
            if not self._size_is_array:
                raise ValueError(
                    "Expected size to be a scalar. Use with_size_multi() to "
                    "set multiple sizes."
                )
            size = as_array_1d(size)
            if size.size != ndata:
                raise ValueError(
                    f"Expected `size` to have the same size as the layer data size "
                    f"({self.ndata}), got {size.size}."
                )
        if ndata > 0:
            self._backend._plt_set_symbol_size(size)
        self.events.size.emit(size)

    def update(
        self,
        *,
        symbol: Symbol | str | _Void = _void,
        size: float | _Void = _void,
        color: ColorType | _Void = _void,
        alpha: float | _Void = _void,
        hatch: str | Hatch | _Void = _void,
    ) -> Self:
        """Update the properties of the markers."""
        if symbol is not _void:
            self.symbol = symbol
        if size is not _void:
            self.size = size
        if color is not _void:
            self.face.color = color
        if hatch is not _void:
            self.face.hatch = hatch
        if alpha is not _void:
            self.face.alpha = alpha
        return self

    def color_by_density(
        self,
        cmap: ColormapType = "jet",
        *,
        width: float = 0.0,
    ) -> Self:
        """
        Set the color of the markers by density.

        Parameters
        ----------
        cmap : ColormapType, optional
            Colormap used to map the density to colors.
        """
        from whitecanvas.utils.kde import gaussian_kde

        xydata = self.data
        xy = np.vstack([xydata.x, xydata.y])
        density = gaussian_kde(xy)(xy)
        normed = density / density.max()
        self.with_face_multi(color=Colormap(cmap)(normed))
        if width is not None:
            self.width = width
        return self

    def as_edge_only(
        self,
        *,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Convert the markers to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        color = self.face.color
        if color.ndim == 0:
            pass
        elif color.ndim == 1:
            self.with_edge(color=color, width=width, style=style)
        elif color.ndim == 2:
            self.with_edge_multi(color=color, width=width, style=style)
        else:
            raise RuntimeError("Unreachable error.")
        self.face.update(alpha=0.0)
        return self

    def with_hover_template(
        self,
        template: str,
        extra: Any | None = None,
    ) -> Self:
        """Add hover template to the markers."""
        xs, ys = self.data
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, xs.size, extra)
        # set default format keys
        params.setdefault("x", xs)
        params.setdefault("y", ys)
        if "i" not in params:
            params["i"] = np.arange(xs.size)
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(xs.size)
        ]
        self._backend._plt_set_hover_text(texts)
        return self

    def with_xerr(
        self,
        err: ArrayLike,
        err_high: ArrayLike | None = None,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = True,
        capsize: float = 0,
    ) -> _lg.LabeledMarkers:
        """
        Add horizontal error bars to the markers.

        Parameters
        ----------
        err : ArrayLike
            Error values. If `err_high` is not specified, the error bars are
            symmetric.
        err_high : array-like, optional
            Upper error values.
        color : color-like, optional
            Line color of the error bars.
        width : float
            Width of the error bars.
        style : str or LineStyle
            Line style of the error bars.
        antialias : bool
            Antialiasing of the error bars.
        capsize : float, optional
            Size of the caps at the end of the error bars.

        Returns
        -------
        LabeledMarkers
            Layer group containing the markers and the error bars as children.
        """
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group import LabeledMarkers

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.edge.color
        if width is _void:
            width = self.edge.width
        if style is _void:
            style = self.edge.style
        xerr = Errorbars(
            self.data.y, self.data.x - err, self.data.x + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            orient=Orientation.HORIZONTAL, backend=self._backend_name
        )  # fmt: skip
        yerr = Errorbars.empty_v(f"xerr-of-{self.name}", backend=self._backend_name)
        return LabeledMarkers(self, xerr, yerr, name=self.name)

    def with_yerr(
        self,
        err: ArrayLike,
        err_high: ArrayLike | None = None,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool = True,
        capsize: float = 0,
    ) -> _lg.LabeledMarkers:
        """
        Add vertical error bars to the markers.

        Parameters
        ----------
        err : ArrayLike
            Error values. If `err_high` is not specified, the error bars are
            symmetric.
        err_high : array-like, optional
            Upper error values.
        color : color-like, optional
            Line color of the error bars.
        width : float
            Width of the error bars.
        style : str or LineStyle
            Line style of the error bars.
        antialias : bool
            Antialiasing of the error bars.
        capsize : float, optional
            Size of the caps at the end of the error bars.

        Returns
        -------
        LabeledMarkers
            Layer group containing the markers and the error bars as children.
        """
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group import LabeledMarkers

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.edge.color
        if width is _void:
            width = self.edge.width
        if style is _void:
            style = self.edge.style
        yerr = Errorbars(
            self.data.x, self.data.y - err, self.data.y + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            orient=Orientation.VERTICAL, backend=self._backend_name
        )  # fmt: skip
        xerr = Errorbars.empty_h(f"yerr-of-{self.name}", backend=self._backend_name)
        return LabeledMarkers(self, xerr, yerr, name=self.name)

    def with_text(
        self,
        strings: str | list[str],
        *,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        fontfamily: str | None = None,
    ) -> _lg.LabeledMarkers:
        """
        Add texts at the positions of the data points.

        Parameters
        ----------
        strings : str, sequence of str
            String values to be added. If single str is given, same string will be added
            to all of the positions.
        color : ColorType, optional
            Color of the text string.
        size : float, default 12
            Point size of the text.
        rotation : float, default 0.0
            Rotation angle of the text in degrees.
        anchor : str or Alignment, default Alignment.BOTTOM_LEFT
            Anchor position of the text. The anchor position will be the coordinate
            given by (x, y).
        family : str, optional
            Font family of the text.

        Returns
        -------
        LabeledMarkers
            Layer group containing the markers.
        """
        from whitecanvas.layers import Errorbars
        from whitecanvas.layers.group import LabeledMarkers

        old_name = self.name
        strings = _text_utils.norm_label_text(strings, self.data)
        texts = Texts(
            *self.data, strings, color=color, size=size, rotation=rotation,
            name=f"text-of-{old_name}", anchor=anchor, family=fontfamily,
            backend=self._backend_name,
        )  # fmt: skip
        self.name = f"markers-of-{old_name}"
        return LabeledMarkers(
            self,
            Errorbars.empty_h(f"xerr-of-{old_name}", backend=self._backend_name),
            Errorbars.empty_v(f"yerr-of-{old_name}", backend=self._backend_name),
            texts=texts,
            name=old_name,
        )

    def with_network(
        self,
        connections: NDArray[np.intp],
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool = True,
    ) -> _lg.Graph:
        """
        Add network edges to the markers to create a graph.

        Parameters
        ----------
        connections : (N, 2) array of int
            Integer array that defines the connections between nodes.
        color : color-like, optional
            Color of the lines.
        width : float, optional
            Width of the line.
        style : str, optional
            Line style of the line.
        antialias : bool, optional
            Antialiasing of the line.

        Returns
        -------
        Graph
            A Graph layer that contains the markers and the edges as children.
        """
        from whitecanvas.layers._primitive import MultiLine
        from whitecanvas.layers.group import Graph

        edges = np.asarray(connections, dtype=np.intp)
        if edges.ndim != 2 or edges.shape[1] != 2:
            raise ValueError("edges must be a (N, 2) array")

        if color is _void:
            color = self.edge.color
        if width is _void:
            width = self.edge.width
        if style is _void:
            style = self.edge.style
        segs = []
        nodes = self.data.stack()
        for i0, i1 in edges:
            segs.append(np.stack([nodes[i0], nodes[i1]], axis=0))
        edges_layer = MultiLine(
            segs, name="edges", color=color, width=width, style=style,
            antialias=antialias, backend=self._backend_name
        )  # fmt: skip
        texts = Texts(
            nodes[:, 0], nodes[:, 1], [""] * nodes.shape[0], name="texts",
            backend=self._backend_name,
        )  # fmt: skip
        return Graph(self, edges_layer, texts, edges, name=self.name)

    def with_stem(
        self,
        orient: OrientationLike = "vertical",
        *,
        bottom: NDArray[np.floating] | float | None = None,
        color: ColorType | _Void = _void,
        alpha: float = 1.0,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool = True,
    ) -> _lg.StemPlot:
        """
        Grow stems from the markers.

        Parameters
        ----------
        orient : str or Orientation, default "vertical"
            Orientation to grow stems.
        bottom : float or array-like, optional
            Bottom of the stems. If not specified, the bottom is set to 0.
        color : color-like, optional
            Color of the lines.
        alpha : float, optional
            Alpha channel of the lines.
        width : float, optional
            Width of the lines.
        style : str or LineStyle
            Line style used to draw the stems.
        antialias : bool, optional
            Line antialiasing.

        Returns
        -------
        StemPlot
            StemPlot layer containing the markers and the stems as children.
        """
        from whitecanvas.layers._primitive import MultiLine
        from whitecanvas.layers.group import StemPlot

        ori = Orientation.parse(orient)
        xdata, ydata = self.data
        if bottom is None:
            bottom = np.zeros_like(ydata)
        elif isinstance(bottom, (float, int, np.number)):
            bottom = np.full((ydata.size,), bottom)
        else:
            bottom = as_array_1d(bottom)
            if bottom.shape != ydata.shape:
                raise ValueError(
                    "Expected bottom to have the same size as ydata, "
                    f"got {bottom.shape} and {ydata.shape}"
                )
        if ori.is_vertical:
            root = np.stack([xdata, bottom], axis=1)
            leaf = np.stack([xdata, ydata], axis=1)
        else:
            root = np.stack([bottom, ydata], axis=1)
            leaf = np.stack([xdata, ydata], axis=1)
        segs = np.stack([root, leaf], axis=1)
        if color is _void:
            color = self.edge.color
        if width is _void:
            width = self.edge.width
        if style is _void:
            style = self.edge.style
        mline = MultiLine(
            segs, name="stems", color=color, width=width, style=style,
            antialias=antialias, alpha = alpha, backend=self._backend_name,
        )  # fmt: skip
        return StemPlot(self, mline, orient=orient, name=self.name)

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1.0,
    ) -> Markers[ConstFace, _Edge, _Size]:
        """
        Return a markers layer with constant face properties.

        In most cases, this function is not needed because the face properties
        can be set directly on construction. Examples below are equivalent:

        >>> markers = canvas.add_markers(x, y).with_face(color="red")
        >>> markers = canvas.add_markers(x, y, color="red")

        This function will be useful when you want to change the markers from
        multi-face state to constant-face state:

        >>> markers = canvas.add_markers(x, y).with_face_multi(color=colors)
        >>> markers = markers.with_face(color="red")

        Parameters
        ----------
        color : color-like, optional
            Color of the marker faces.
        hatch : str or FacePattern, optional
            Hatch pattern of the faces.
        alpha : float, optional
            Alpha channel of the faces.

        Returns
        -------
        Markers
            The updated markers layer.
        """
        super().with_face(color, hatch, alpha)
        if not self.symbol.has_face():
            width = self.edge.width
            if isinstance(width, (float, int, np.number)):
                self.edge.update(width=width or 1.0, color=color)
            else:
                self.edge.update(width=width[0] or 1.0, color=color)
        return self

    def with_face_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] | _Void = _void,
        alpha: float = 1,
    ) -> Markers[MultiFace, _Edge, _Size]:
        """
        Return a markers layer with multi-face properties.

        This function is used to create a markers layer with multiple face
        properties, such as colorful markers.

        >>> markers = canvas.add_markers(x, y).with_face_multi(color=colors)

        Parameters
        ----------
        color : color-like or sequence of color-like, optional
            Color(s) of the marker faces.
        hatch : str or Hatch or sequence of it, optional
            Pattern(s) of the faces.
        alpha : float or sequence of float, optional
            Alpha channel(s) of the faces.

        Returns
        -------
        Markers
            The updated markers layer.
        """
        super().with_face_multi(color, hatch, alpha)
        if not self.symbol.has_face():
            width = self.edge.width
            if isinstance(width, (float, int, np.number)):
                self.edge.update(width=width or 1.0, color=color)
            else:
                self.edge.update(width=width[0] or 1.0, color=color)
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1.0,
    ) -> Markers[_Face, ConstEdge, _Size]:
        return super().with_edge(color, width, style, alpha)

    def with_edge_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1.0,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1.0,
    ) -> Markers[_Face, MultiEdge, _Size]:
        """
        Return a markers layer with multi-edge properties.

        This function is used to create a markers layer with multiple edge
        properties, such as colorful markers.

        >>> markers = canvas.add_markers(x, y).with_edge_multi(color=colors)

        Parameters
        ----------
        color : color-like or sequence of color-like, optional
            Color(s) of the marker faces.
        width : float or array of float, optional
            Width(s) of the edges.
        style : str, LineStyle or sequence of them, optional
            Line style(s) of the edges.
        alpha : float or sequence of float, optional
            Alpha channel(s) of the faces.

        Returns
        -------
        Markers
            The updated markers layer.
        """
        return super().with_edge_multi(color, width, style, alpha)

    def with_size_multi(
        self,
        size: float | Sequence[float],
    ) -> Markers[_Face, _Edge, NDArray[np.float32]]:
        if isinstance(size, (float, int, np.number)):
            _size = np.full(self.ndata, size, dtype=np.float32)
        else:
            _size = as_array_1d(size, dtype=np.float32)
            if _size.size != self.ndata:
                raise ValueError(
                    "Expected size to have the same size as the data, "
                    f"got {_size.size} and {self.ndata}"
                )
        self._size_is_array = True
        self.size = _size
        return self

    def _as_all_multi(self) -> Markers[MultiFace, MultiEdge, NDArray[np.float32]]:
        return (
            self.with_face_multi().with_edge_multi(width=0.0).with_size_multi(self.size)
        )

    def _as_legend_item(self) -> _legend.MarkersLegendItem:
        if self._size_is_array:
            size = np.mean(self.size)
        else:
            size = self.size
        return _legend.MarkersLegendItem(
            self.symbol, size, self.face._as_legend_info(), self.edge._as_legend_info()
        )



from __future__ import annotations

from typing import Any

import numpy as np

from whitecanvas.backend import Backend
from whitecanvas.layers._primitive.line import LineLayerEvents, LineMixin
from whitecanvas.layers._sizehint import xyz_size_hint
from whitecanvas.layers.layer3d._base import DataBoundLayer3D
from whitecanvas.protocols import LineProtocol
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    LineStyle,
    XYZData,
)
from whitecanvas.utils.normalize import as_array_1d, normalize_xyz


class Line3D(LineMixin[LineProtocol], DataBoundLayer3D[LineProtocol, XYZData]):
    _backend_class_name = "components3d.MonoLine3D"
    events: LineLayerEvents
    _events_class = LineLayerEvents

    def __init__(
        self,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        zdata: ArrayLike1D,
        *,
        name: str | None = None,
        color: ColorType = "blue",
        width: float = 1,
        alpha: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        xdata, ydata, zdata = normalize_xyz(xdata, ydata, zdata)
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), xdata, ydata, zdata)
        self.update(
            color=color, width=width, style=style, alpha=alpha, antialias=antialias
        )
        self._x_hint, self._y_hint, self._z_hint = xyz_size_hint(xdata, ydata, zdata)
        # self._backend._plt_connect_pick_event(self.events.clicked.emit)

    def _get_layer_data(self) -> XYZData:
        return XYZData(*self._backend._plt_get_data())

    def _norm_layer_data(self, data: Any) -> XYZData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 3:
                raise ValueError(f"Expected data to be (N, 3), got {data.shape}")
            xdata, ydata, zdata = data[:, 0], data[:, 1], data[:, 2]
        else:
            xdata, ydata, zdata = data
            if xdata is None:
                xdata = self.data.x
            else:
                xdata = as_array_1d(xdata)
            if ydata is None:
                ydata = self.data.y
            else:
                ydata = as_array_1d(ydata)
            if zdata is None:
                zdata = as_array_1d(zdata)
        if xdata.size != ydata.size or xdata.size != zdata.size:
            raise ValueError(
                "Expected xdata, ydata and zdata to have the same size, "
                f"got {xdata.size}, {ydata.size} and {zdata.size}."
            )
        return XYZData(xdata, ydata)

    def _set_layer_data(self, data: XYZData):
        x0, y0, z0 = data
        self._backend._plt_set_data(x0, y0, z0)
        self._x_hint, self._y_hint, self._z_hint = xyz_size_hint(x0, y0, z0)

    def set_data(
        self,
        xdata: ArrayLike1D | None = None,
        ydata: ArrayLike1D | None = None,
        zdata: ArrayLike1D | None = None,
    ):
        self.data = xdata, ydata, zdata

    @property
    def ndata(self) -> int:
        """Number of data points."""
        return self.data.x.size

    # def with_markers(
    #     self,
    #     symbol: Symbol | str = Symbol.CIRCLE,
    #     size: float | None = None,
    #     color: ColorType | _Void = _void,
    #     alpha: float = 1.0,
    #     hatch: str | Hatch = Hatch.SOLID,
    # ) -> _lg.Plot:



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from cmap import Colormap
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers._base import HoverableDataBoundLayer
from whitecanvas.layers._primitive.line import (
    MultiLine,
    MultiLineEvents,
    MultiLineProtocol,
)
from whitecanvas.types import (
    ArrayLike1D,
    ColormapType,
    ColorType,
    KdeBandWidthType,
    LineStyle,
    Orientation,
    OrientationLike,
    XYYData,
)
from whitecanvas.utils.normalize import as_array_1d
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self


class Rug(MultiLine, HoverableDataBoundLayer[MultiLineProtocol, NDArray[np.number]]):
    """
    Rug plot (event plot) layer.

      │ ││  │   │
    ──┴─┴┴──┴───┴──>
    """

    _ATTACH_TO_AXIS = True
    events: MultiLineEvents
    _events_class = MultiLineEvents

    def __init__(
        self,
        events: ArrayLike1D,
        *,
        low: ArrayLike1D | float = 0.0,
        high: ArrayLike1D | float = 1.0,
        name: str | None = None,
        color: ColorType = "black",
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        alpha: float = 1.0,
        antialias: bool = True,
        orient: OrientationLike = "vertical",
        backend: str | Backend | None = None,
    ):
        events, segs, orient = _norm_input(events, low, high, orient)
        super().__init__(
            segs, name=name, color=color, alpha=alpha, width=width, style=style,
            antialias=antialias, backend=backend
        )  # fmt: skip
        self._orient = Orientation.parse(orient)

    def _get_layer_data(self) -> NDArray[np.number]:
        segs = super()._get_layer_data()
        idx = 0 if self.orient.is_vertical else 1
        return np.array([seg[0, idx] for seg in segs])

    def _norm_layer_data(self, data: Any) -> NDArray[np.number]:
        return as_array_1d(data)

    def _set_layer_data(self, data: NDArray[np.number]):
        _, segs, _ = _norm_input(data, self.low, self.high, self.orient)
        return super()._set_layer_data(segs)

    @property
    def low(self) -> NDArray[np.number]:
        """Coordinate of the lower bound."""
        return self.data_full.y0

    @low.setter
    def low(self, val):
        _, segs, _ = _norm_input(self.data, val, self.high, self.orient)
        super()._set_layer_data(segs)
        self.events.data.emit(segs)

    @property
    def high(self) -> NDArray[np.number]:
        """Coordinate of the higher bound."""
        return self.data_full.y1

    @high.setter
    def high(self, val):
        _, segs, _ = _norm_input(self.data, self.low, val, self.orient)
        super()._set_layer_data(segs)
        self.events.data.emit(segs)

    @property
    def data_full(self) -> XYYData:
        segs = MultiLine._get_layer_data(self)
        seg_stack = np.stack(segs, axis=2)
        idx = 1 if self.orient.is_vertical else 0
        low = seg_stack[0, idx, :]
        high = seg_stack[1, idx, :]
        _t = seg_stack[0, 1 - idx, :]
        return XYYData(x=_t, y0=low, y1=high)

    @data_full.setter
    def data_full(self, val: XYYData):
        val = XYYData(*val)
        _, segs, _ = _norm_input(val.x, val.y0, val.y1, self.orient)
        super()._set_layer_data(segs)
        self.events.data.emit(segs)

    def set_data(self, events: ArrayLike1D):
        self.data = events

    @property
    def orient(self) -> Orientation:
        """Orientation of the rug plot."""
        return self._orient

    def update_length(
        self,
        lengths: float | NDArray[np.number],
        *,
        offset: float | None = None,
        align: str = "low",
    ) -> Self:
        """
        Update the length of the rug lines.

        Parameters
        ----------
        lengths : float or array-like
            Length of the rug lines. If a scalar, all the lines have the same length.
            If an array, each line has a different length.
        offset : float, optional
            Offset of the lines. If not given, the mean of the lower and upper bounds is
            used.
        align : {'low', 'high', 'center'}, optional
            How to align the rug lines around the offset. This parameter is defined as
            follows.

            ```
               "low"     "high"    "center"
              ──┴─┴──   ──┬─┬──    ──┼─┼──
            ```
        """
        data_full = self.data_full
        if offset is None:
            offset = data_full.y0.mean()
        if not is_real_number(lengths):
            lengths = as_array_1d(lengths)
            if lengths.size != data_full.x.size:
                raise ValueError(
                    "`lengths` must be a scalar or an array with the same size as "
                    f"events, but got {lengths.size} and {data_full.x.size} events."
                )

        if align == "low":
            y0 = np.full((data_full.x.size,), offset)
            y1 = offset + lengths
        elif align == "high":
            y0 = offset - lengths
            y1 = np.full((data_full.x.size,), offset)
        elif align == "center":
            y0 = offset - lengths / 2
            y1 = offset + lengths / 2
        else:
            raise ValueError(
                f"`align` must be 'low', 'high', or 'center', got {align!r}."
            )
        self.data_full = XYYData(data_full.x, y0, y1)
        return self

    def color_by_density(
        self,
        cmap: ColormapType = "jet",
        *,
        band_width: KdeBandWidthType = "scott",
    ) -> Self:
        """
        Set the color of the markers by density.

        Parameters
        ----------
        cmap : ColormapType, optional
            Colormap used to map the density to colors.
        band_width : float, "scott" or "silverman", optional
            Method to calculate the estimator bandwidth.
        """
        from whitecanvas.utils.kde import gaussian_kde

        events = self.data
        density = gaussian_kde(events, band_width)(events)
        normed = density / density.max()
        self.color = Colormap(cmap)(normed)
        return self

    def scale_by_density(
        self,
        max_length: float = 1.0,
        *,
        offset: float | None = None,
        align: str = "low",
        band_width: KdeBandWidthType = "scott",
    ) -> Self:
        """
        Set the height of the lines by density.

        Parameters
        ----------
        max_length : float, optional
            Maximum length of the lines.
        offset : float, optional
            Offset of the lines. If not given, the mean of the lower and upper bounds is
            used.
        align : {'low', 'high', 'center'}, optional
            How to align the rug lines around the offset. This parameter is defined as
            follows.

            ```
               "low"     "high"    "center"
              ──┴─┴──   ──┬─┬──    ──┼─┼──
            ```
        band_width : float, "scott" or "silverman", optional
            Method to calculate the estimator bandwidth.
        """
        from whitecanvas.utils.kde import gaussian_kde

        events = self.data
        density = gaussian_kde(events, band_width)(events)
        normed = density / density.max() * max_length
        return self.update_length(normed, offset=offset, align=align)


def _norm_input(
    events: ArrayLike1D,
    low: ArrayLike1D | float,
    high: ArrayLike1D | float,
    orient: OrientationLike,
) -> tuple[NDArray[np.number], NDArray[np.number], Orientation]:
    _t = as_array_1d(events)
    if is_real_number(low):
        y0 = np.full(_t.shape, low)
    else:
        y0 = as_array_1d(low)
        if y0.size != _t.size:
            raise ValueError(
                "`low` must be a scalar or an array with the same size as events, but "
                f"got {y0.size} and {events.size} events."
            )
    if is_real_number(high):
        y1 = np.full(_t.shape, high)
    else:
        y1 = as_array_1d(high)
        if y1.size != _t.size:
            raise ValueError(
                "`high` must be a scalar or an array with the same size as events, but "
                f"got {y1.size} and {events.size} events."
            )
    ori = Orientation.parse(orient)
    if ori.is_vertical:
        start = np.stack([_t, y0], axis=1)
        stop = np.stack([_t, y1], axis=1)
        segs = np.stack([start, stop], axis=0)
    else:
        start = np.stack([y0, _t], axis=1)
        stop = np.stack([y1, _t], axis=1)
        segs = np.stack([start, stop], axis=0)
    return events, np.moveaxis(segs, 1, 0), orient



from __future__ import annotations

import warnings
import weakref
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

from plotly import graph_objs as go

from whitecanvas.types import LineStyle, Symbol

if TYPE_CHECKING:
    from plotly.basedatatypes import BaseTraceType

_O = TypeVar("_O", bound="BaseTraceType")


class PlotlyLayer(Generic[_O]):
    _props: dict[str, Any] | _O
    _fig_ref: Callable[[], go.Figure]

    def _plt_get_visible(self) -> bool:
        return self._props["visible"]

    def _plt_set_visible(self, visible: bool) -> bool:
        self._props["visible"] = visible


class PlotlyHoverableLayer(PlotlyLayer[_O]):
    def __init__(self):
        self._hover_texts: list[str] | None = None
        self._click_callbacks = []
        self._fig_ref = lambda: None

    def _plt_connect_pick_event(self, callback):
        fig = self._fig_ref()
        if fig is None:
            self._click_callbacks.append(callback)
        else:
            self._connect_mouse_events(fig, [callback])

    def _connect_mouse_events(
        self,
        fig: go.Figure,
        callbacks: list[Callable] | None = None,
    ):
        if callbacks is None:
            callbacks = self._click_callbacks
        gobj = self._props
        if not hasattr(gobj, "uid"):
            raise ValueError("Graph object is not created yet.")
        for cb in callbacks:
            gobj.on_click(_convert_cb(cb), append=True)
        self._fig_ref = weakref.ref(fig)
        self._update_hover_texts(fig)

    def _plt_set_hover_text(self, text: list[str]):
        self._hover_texts = text
        fig = self._fig_ref()
        if fig is not None:
            self._update_hover_texts(fig)

    def _update_hover_texts(self, fig: go.Figure):
        if self._hover_texts is None:
            return
        if len(self._hover_texts) != self._plt_get_ndata():
            warnings.warn(
                f"Length of hover text ({len(self._hover_texts)}) does not match the "
                f"number of data points ({self._plt_get_ndata()}). Ignore updating.",
                RuntimeWarning,
                stacklevel=2,
            )
            return

        fig.update_traces(
            customdata=self._hover_texts,
            selector={"uid": self._props["uid"]},
        )


@dataclass
class FigureLocation:
    row: int
    col: int
    secondary_y: bool = False

    def asdict(self):
        out = {}
        if self.row > 1 or self.col > 1:
            out["row"] = self.row
            out["col"] = self.col
        if self.secondary_y:
            out["secondary_y"] = True
        return out

    def asdictn(self):
        out = {}
        if self.row > 1 or self.col > 1:
            out["rows"] = self.row
            out["cols"] = self.col
        if self.secondary_y:
            out["secondary_y"] = True
        return out


_LINE_STYLES = {
    "solid": LineStyle.SOLID,
    "dash": LineStyle.DASH,
    "dashdot": LineStyle.DASH_DOT,
    "dot": LineStyle.DOT,
}


def from_plotly_linestyle(ls: str) -> LineStyle:
    return _LINE_STYLES.get(ls, LineStyle.SOLID)


def to_plotly_linestyle(ls: LineStyle) -> str:
    return ls.name.lower().replace("_", "")


def from_plotly_marker_symbol(s: str) -> Symbol:
    return _SYMBOLS.get(s, Symbol.CIRCLE)


def to_plotly_marker_symbol(s: Symbol) -> str:
    return _SYMBOLS_INV.get(s, "circle")


_SYMBOLS = {
    "circle": Symbol.CIRCLE,
    "square": Symbol.SQUARE,
    "diamond": Symbol.DIAMOND,
    "cross": Symbol.PLUS,
    "x": Symbol.CROSS,
    "triangle-up": Symbol.TRIANGLE_UP,
    "triangle-down": Symbol.TRIANGLE_DOWN,
    "triangle-left": Symbol.TRIANGLE_LEFT,
    "triangle-right": Symbol.TRIANGLE_RIGHT,
    "star": Symbol.STAR,
    "line-ew": Symbol.HBAR,
    "line-ns": Symbol.VBAR,
}

_SYMBOLS_INV = {v: k for k, v in _SYMBOLS.items()}


def _convert_cb(cb):
    def _out(trace, points, state):
        indices = points.point_inds
        if indices:
            cb(indices)

    return _out


def to_html(canvas) -> str:
    from plotly.io import to_html

    return to_html(canvas._figs)



import numpy as np
from mpl_toolkits.mplot3d import art3d

from whitecanvas.backend.matplotlib._base import MplLayer
from whitecanvas.types import Hatch


class Mesh3D(art3d.Poly3DCollection, MplLayer):
    def __init__(self, verts, faces):
        super().__init__(verts[faces])
        self._verts = verts
        self._faces = faces

    def _plt_get_data(self):
        return self._verts, self._faces

    def _plt_set_data(self, verts, faces):
        self._verts = verts
        self._faces = faces
        self.set_verts(verts[faces])

    def _plt_get_face_color(self):
        if self.axes is None:
            return self._facecolor3d
        return self.get_facecolor()

    def _plt_set_face_color(self, color):
        if self.axes is None:
            self._facecolor3d = color
        else:
            self.set_facecolor(color)

    def _plt_get_face_hatch(self):
        return self.get_hatch()

    def _plt_set_face_hatch(self, hatch):
        if isinstance(hatch, Hatch):
            self.set_hatch(hatch.value)
        else:
            self.set_hatch(hatch[0])

    def _plt_get_edge_color(self):
        if self.axes is None:
            return np.array([0, 0, 0, 0])
        return self.get_edgecolor()[0]

    def _plt_set_edge_color(self, color):
        if self.axes is None:
            self._edgecolor3d = color
        else:
            self.set_edgecolor(color)

    def _plt_get_edge_width(self):
        return self.get_linewidth()[0]

    def _plt_set_edge_width(self, width):
        self.set_linewidth(width)

    def _plt_get_edge_style(self):
        return self.get_linestyle()

    def _plt_set_edge_style(self, style):
        self.set_linestyle(style)

    def post_add(self, canvas):
        self._plt_set_face_color(self._facecolor3d)
        self._plt_set_edge_color(self._edgecolor3d)



from __future__ import annotations

import weakref
from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    Iterator,
    Literal,
    Sequence,
    TypeVar,
    Union,
)

import numpy as np

from whitecanvas._exceptions import ReferenceDeletedError
from whitecanvas.layers.tabular import parse
from whitecanvas.utils.collections import OrderedSet

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers.tabular._dataframe import DataFrameWrapper

_C = TypeVar("_C")  # NOTE: don't have to be a canvas
_T = TypeVar("_T")
_DF = TypeVar("_DF")
NStr = Union[str, Sequence[str]]
AggMethods = Literal["min", "max", "mean", "median", "sum", "std"]


class BaseCatPlotter(Generic[_C, _DF]):
    def __init__(
        self,
        canvas: _C,
        df: _DF,
    ):
        self._canvas_ref = weakref.ref(canvas)
        self._df: DataFrameWrapper[_DF] = parse(df)

    def _canvas(self) -> _C:
        canvas = self._canvas_ref()
        if canvas is None:
            raise ReferenceDeletedError("Canvas has been deleted.")
        return canvas

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args) -> None:
        pass


class CatIterator(Generic[_DF]):
    def __init__(
        self,
        df: DataFrameWrapper[_DF],
        offsets: tuple[str, ...],
        numeric: bool = False,
        sort_func: Callable[[Sequence[_T]], Sequence[_T]] | None = None,
    ):
        self._df = df
        self._offsets = tuple(offsets)
        self._cat_map_cache = {}
        self._numeric = numeric
        if sort_func is None:
            sort_func = lambda x: x  # noqa: E731
        elif numeric:
            raise ValueError("sort_func is not allowed for numeric data.")
        self._sort_func = sort_func

    @property
    def df(self) -> DataFrameWrapper[_DF]:
        return self._df

    @property
    def offsets(self) -> tuple[str, ...]:
        return self._offsets

    def category_map(self, columns: tuple[str, ...]) -> dict[tuple, int]:
        """Calculate how to map category columns to integers."""
        columns = tuple(columns)
        if columns in self._cat_map_cache:
            return self._cat_map_cache[columns]
        if len(columns) == 0:
            return {(): 0}
        serieses = [self._df[c] for c in columns]
        if columns == self._offsets:
            categories = self._sort_func(OrderedSet(zip(*serieses)))
        else:
            categories = OrderedSet(zip(*serieses))
        _map = {uni: i for i, uni in enumerate(categories)}
        self._cat_map_cache[columns] = _map
        return _map

    def iter_arrays(
        self,
        by: tuple[str, ...],
        dodge: tuple[str, ...] | None = None,
        width: float = 0.8,
    ) -> Iterator[tuple[tuple, float, DataFrameWrapper[_DF]]]:
        """
        Iterate over the groups of the DataFrame for plotting.

        Returns
        -------
        Iterator of (tuple, float, DataFrameWrapper[_DF])
            The first tuple is the group key, the second float is the x (or y)
            coordinate the group should be plotted at, and the third is the subset
            of the DataFrame that corresponds to the group.
        """
        if dodge is None:
            dodge = ()
        if set(self._offsets) > set(by):
            raise ValueError(
                f"offsets must be a subset of by, got offsets={self._offsets!r} and "
                f"by={by!r}"
            )
        indices = [by.index(d) for d in self._offsets]
        if self._numeric:
            _map = NumericMap()
        else:
            _map = self.category_map(self._offsets)
        if not dodge:
            for sl, group in self._df.group_by(by):
                key = tuple(sl[i] for i in indices)
                yield sl, _map[key], group
        else:
            if set(self._offsets) & set(dodge):
                raise ValueError(
                    f"offsets and dodge must be disjoint, got offsets={self._offsets!r}"
                    f" and dodge={dodge!r}"
                )
            if self._numeric:
                _pos = list(self.prep_position_map(self._offsets, dodge=False).values())
                _width = np.diff(np.sort(_pos)).min() * width
            else:
                _width = width
            inv_indices = [by.index(d) for d in dodge]
            _res_map = self.category_map(dodge)
            _nres = len(_res_map)
            dmax = (_nres - 1) / 2 / _nres * _width
            dd = np.linspace(-dmax, dmax, _nres)
            for sl, group in self._df.group_by(by):
                key = tuple(sl[i] for i in indices)
                res = tuple(sl[i] for i in inv_indices)
                yield sl, dd[_res_map[res]] + _map[key], group

    def prep_arrays(
        self,
        by: tuple[str, ...],
        value: str,
        dodge: tuple[str, ...] | None = None,
        width: float = 0.8,
    ) -> tuple[list[float], list[np.ndarray], list[tuple]]:
        x = []
        arrays = []
        categories = []
        for sl, offset, group in self.iter_arrays(by, dodge, width=width):
            x.append(offset)
            arrays.append(group[value])
            categories.append(sl)
        return x, arrays, categories

    def prep_position_map(
        self,
        by: tuple[str],
        dodge: tuple[str, ...] | None = None,
        width: float = 0.8,
    ) -> dict[tuple, float]:
        out = {}
        for sl, offset, _ in self.iter_arrays(by, dodge=dodge, width=width):
            out[sl] = offset
        return out

    def axis_ticks(self) -> tuple[list[float], list[str]]:
        """Prepare the axis ticks and labels for the category plot."""
        pos: list[float] = []
        labels: list[str] = []
        for k, v in self.category_map(self._offsets).items():
            pos.append(v)
            labels.append("\n".join(map(str, k)))
        pos_indices = np.argsort(pos)
        return [pos[i] for i in pos_indices], [labels[i] for i in pos_indices]

    def axis_label(self) -> str:
        return "/".join(self._offsets)

    def zoom_factor(
        self,
        dodge: tuple[str, ...] | None = None,
        width: float = 0.8,
    ) -> float:
        """Return the zoom factor for the given dodge."""
        if dodge:
            _res_map = self.category_map(dodge)
            _nres = len(_res_map)
            if _nres == 1:
                return 1.0
            dmax = (_nres - 1) / 2 / _nres * width
            return 2 * dmax / (_nres - 1)
        else:
            return 1.0

    def categories(self) -> list[tuple]:
        return list(self.category_map(self._offsets).keys())


class NumericMap:
    def __getitem__(self, key: tuple[float]) -> float:
        return key[0]



from __future__ import annotations

from typing import Generic, TypeVar, overload

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._primitive import Band, Line
from whitecanvas.layers._primitive.line import _SingleLine
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    HistBinType,
    HistogramKind,
    HistogramShape,
    KdeBandWidthType,
    LineStyle,
    Orientation,
    OrientationLike,
    XYData,
)
from whitecanvas.utils.hist import get_hist_edges, histograms
from whitecanvas.utils.normalize import as_array_1d

_L = TypeVar("_L", bound=_SingleLine)


class LineFillBase(LayerContainer, Generic[_L]):
    _ATTACH_TO_AXIS = True

    def __init__(self, line: _L, fill: Band, name: str | None = None):
        super().__init__([line, fill], name=name)
        self._fill_alpha = 0.2

    @property
    def line(self) -> _L:
        """The line layer."""
        return self._children[0]

    @property
    def fill(self) -> Band:
        """The fill layer."""
        return self._children[1]

    @property
    def orient(self) -> Orientation:
        """Orientation of the line and fill layers."""
        return self.fill.orient

    @property
    def color(self) -> NDArray[np.float32]:
        """Color of the layer."""
        return self.line.color

    @color.setter
    def color(self, color: ColorType):
        self.line.color = color
        self.fill.face.update(color=color, alpha=self._fill_alpha)
        self.fill.edge.update(color=color, alpha=1.0)

    @property
    def fill_alpha(self) -> float:
        """The alpha value applied to the fill region compared to the line."""
        return self._fill_alpha

    @fill_alpha.setter
    def fill_alpha(self, alpha: float):
        self._fill_alpha = alpha
        self.fill.face.alpha = alpha
        self.fill.edge.alpha = alpha

    def _as_legend_item(self) -> _legend.BarLegendItem:
        edge = _legend.EdgeInfo(self.line.color, self.line.width, self.line.style)
        return _legend.BarLegendItem(self.fill.face, edge)


class Histogram(LineFillBase[Line]):
    def __init__(
        self,
        data: NDArray[np.number],
        edges: NDArray[np.number],
        limits: tuple[float, float] | None,
        line: Line,
        fill: Band,
        shape: HistogramShape = HistogramShape.bars,
        kind: HistogramKind = HistogramKind.count,
        name: str | None = None,
    ):
        if name is None:
            name = "histogram"
        super().__init__(line, fill, name=name)
        self._data = data
        self._shape = shape
        self._kind = kind
        self._edges = edges
        self._limits = limits

    @property
    def data(self) -> NDArray[np.number]:
        """The data used to plot the histogram."""
        return self._data

    @data.setter
    def data(self, data: NDArray[np.number]):
        data = as_array_1d(data)
        xdata, ydata, _ = self._calculate_xy(
            data, self._edges, self._shape, self._kind, self._limits, clip=True
        )  # fmt: skip
        self._update_internal(xdata, ydata)
        self._data = data

    def _update_internal(self, xdata: NDArray[np.number], ydata: NDArray[np.number]):
        if self.orient.is_vertical:
            self.line.data = xdata, ydata
        else:
            self.line.data = ydata, xdata
        self.fill.data = xdata, _prep_bottom(ydata), ydata

    @property
    def shape(self) -> HistogramShape:
        """The shape of the histogram."""
        return self._shape

    @shape.setter
    def shape(self, shape: str | HistogramShape):
        shape = HistogramShape(shape)
        xdata, ydata, _ = self._calculate_xy(
            self._data, self._edges, shape, self._kind, self._limits
        )  # fmt: skip
        self._update_internal(xdata, ydata)
        self._shape = shape

    @property
    def kind(self) -> HistogramKind:
        """The kind of the histogram."""
        return self._kind

    @kind.setter
    def kind(self, kind: str | HistogramKind):
        kind = HistogramKind(kind)
        xdata, ydata, _ = self._calculate_xy(
            self._data, self._edges, self._shape, kind, self._limits
        )  # fmt: skip
        self._update_internal(xdata, ydata)
        self._kind = kind

    @property
    def limits(self) -> tuple[float, float] | None:
        """The limits of the histogram."""
        return self._limits

    @limits.setter
    def limits(self, limits: tuple[float, float] | None):
        xdata, ydata, _ = self._calculate_xy(
            self._data, self._edges, self._shape, self._kind, limits
        )
        self._update_internal(xdata, ydata)
        self._limits = limits

    @property
    def edges(self) -> NDArray[np.number]:
        """The edges of the histogram."""
        return self._edges

    @edges.setter
    def edges(self, edges: NDArray[np.number]):
        edges = as_array_1d(edges)
        xdata, ydata, _ = self._calculate_xy(
            self._data, edges, self._shape, self._kind, self._limits
        )
        self._update_internal(xdata, ydata)
        self._edges = edges

    @overload
    def update_edges(
        self, bins: HistBinType, limits: tuple[float, float] | None = None
    ): ...

    @overload
    def update_edges(self, edges: NDArray[np.number]): ...

    def update_edges(self, bins, limits=None):
        """
        Update the edges of the histogram.

        >>> hist.update_edges(20, limits=(0, 10))  # uniform bins
        >>> hist.update_edges([0, 2, 3, 5])  # non-uniform bins
        """
        if limits is not None and not isinstance(bins, (int, np.number)):
            raise TypeError("bins must be an integer when limits are specified.")
        edges = get_hist_edges([self._data], bins, limits)
        self.edges = edges

    @classmethod
    def from_array(
        cls,
        data: NDArray[np.number],
        shape: HistogramShape = HistogramShape.bars,
        kind: HistogramKind = HistogramKind.count,
        name: str | None = None,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        color: ColorType = "black",
        style: str | LineStyle = LineStyle.SOLID,
        width: float = 1.0,
        orient: OrientationLike = "vertical",
        backend: str | Backend | None = None,
    ) -> Histogram:
        """Create a histogram from an array."""
        shape = HistogramShape(shape)
        kind = HistogramKind(kind)
        ori = Orientation.parse(orient)
        xdata, ydata, edges = cls._calculate_xy(data, bins, shape, kind, limits)
        if ori.is_vertical:
            line = Line(
                xdata, ydata, color=color, style=style, width=width, backend=backend
            )  # fmt: skip
        else:
            line = Line(
                ydata, xdata, color=color, style=style, width=width, backend=backend
            )
        fill = Band(
            xdata, _prep_bottom(ydata), ydata, color=color, alpha=0.2, orient=ori,
            backend=backend,
        )  # fmt: skip
        return cls(data, edges, limits, line, fill, shape, kind, name=name)

    @staticmethod
    def _calculate_xy(
        data,
        bins: HistBinType,
        shape: HistogramShape,
        kind: HistogramKind,
        limits: tuple[float, float] | None = None,
        clip: bool = True,
    ) -> tuple[NDArray[np.number], NDArray[np.number], NDArray[np.number]]:
        if clip and limits is not None:
            data = np.clip(data, *limits)
        hist = histograms([data], bins, limits)
        shape = HistogramShape(shape)
        kind = HistogramKind(kind)
        if kind is HistogramKind.count:
            heights = hist.counts[0]
        elif kind is HistogramKind.density:
            heights = hist.density()[0]
        elif kind is HistogramKind.probability:
            heights = hist.probability()[0]
        elif kind is HistogramKind.frequency:
            heights = hist.frequency()[0]
        elif kind is HistogramKind.percent:
            heights = hist.percent()[0]
        else:
            raise ValueError(f"Unknown kind {kind!r}.")

        if shape is HistogramShape.step:
            xdata = np.repeat(hist.edges, 2)
            ydata = np.concatenate([[0], np.repeat(heights, 2), [0]])
        elif shape is HistogramShape.polygon:
            centers = hist.centers()
            xdata = np.concatenate([[centers[0]], centers, [centers[-1]]])
            ydata = np.concatenate([[0], heights, [0]])
        elif shape is HistogramShape.bars:
            edges = hist.edges
            xdata = np.repeat(edges, 3)[1:-1]
            ydata = np.zeros_like(xdata)
            ydata[1::3] = ydata[2::3] = heights
        else:
            raise ValueError(f"Unknown shape {shape!r}.")
        return xdata, ydata, hist.edges


def _prep_bottom(ydata: NDArray[np.number]) -> NDArray[np.number]:
    return np.full(ydata.size, 0)


class Kde(LineFillBase[Line]):
    def __init__(
        self,
        data: NDArray[np.number],
        band_width: float,
        line: Line,
        fill: Band,
        name: str | None = None,
        bottom: float = 0.0,
        scale: float = 1.0,
    ):
        if name is None:
            name = "kde"
        super().__init__(line, fill, name=name)
        self._data = data
        self._bottom = bottom
        self._band_width = band_width
        self._scale = scale

    @property
    def data(self) -> NDArray[np.number]:
        """The data used to plot the histogram."""
        return self._data

    @data.setter
    def data(self, data: NDArray[np.number]):
        data = as_array_1d(data)
        xdata, ydata = self._calculate_params(
            data, self._band_width, self._bottom, self._scale
        )  # fmt: skip
        self._update_internal(xdata, ydata, self._bottom)
        self._data = data

    def _update_internal(
        self, xdata: NDArray[np.number], ydata: NDArray[np.number], bottom: float
    ):
        if self.orient.is_vertical:
            self.line.data = xdata, ydata
        else:
            self.line.data = ydata, xdata
        self.fill.data = xdata, np.full(xdata.size, bottom), ydata

    @property
    def band_width(self) -> float:
        """The band width of the kernel density estimation."""
        return self._band_width

    @band_width.setter
    def band_width(self, band_width: float):
        xdata, ydata, bw = self._calculate_params(
            self._data, band_width, self._bottom, self._scale
        )  # fmt: skip
        self._update_internal(xdata, ydata, self._bottom)
        self._band_width = bw

    @property
    def bottom(self) -> float:
        """The bottom value of the fill."""
        return self._bottom

    @bottom.setter
    def bottom(self, bottom: float):
        xdata, ydata, _ = self._calculate_params(
            self._data, self._band_width, bottom, self._scale
        )  # fmt: skip
        self._update_internal(xdata, ydata, bottom)
        self._bottom = bottom

    @property
    def scale(self) -> float:
        """The scale of the kernel density estimation."""
        return self._scale

    @scale.setter
    def scale(self, scale: float):
        xdata, ydata, _ = self._calculate_params(
            self._data, self._band_width, self._bottom, scale
        )  # fmt: skip
        self._update_internal(xdata, ydata, self._bottom)
        self._scale = scale

    @classmethod
    def from_array(
        cls,
        data: ArrayLike1D,
        bottom: float = 0.0,
        scale: float = 1.0,
        *,
        name: str | None = None,
        band_width: KdeBandWidthType = "scott",
        color: ColorType = "blue",
        style: str | LineStyle = LineStyle.SOLID,
        width: float = 1.0,
        orient: OrientationLike = "vertical",
        backend: Backend | str | None = None,
    ):
        data = as_array_1d(data)
        x, y1, bw = cls._calculate_params(data, band_width, bottom, scale)
        ori = Orientation.parse(orient)
        if ori.is_vertical:
            line = Line(x, y1, color=color, style=style, width=width, backend=backend)
        else:
            line = Line(y1, x, color=color, style=style, width=width, backend=backend)
        fill = Band(
            x, np.full(x.size, bottom), y1, color=color, alpha=0.2, orient=ori,
            backend=backend,
        )  # fmt: skip
        return Kde(data, bw, line, fill, name=name, bottom=bottom, scale=scale)

    @staticmethod
    def _calculate_params(
        data: NDArray[np.number],
        band_width: KdeBandWidthType,
        bottom: float = 0.0,
        scale: float = 1.0,
    ) -> tuple[NDArray[np.number], NDArray[np.number], float]:
        from whitecanvas.utils.kde import gaussian_kde

        data = as_array_1d(data)
        kde = gaussian_kde(data, bw_method=band_width)

        sigma = np.sqrt(kde.covariance[0, 0])
        pad = sigma * 2.5
        x = np.linspace(data.min() - pad, data.max() + pad, 100)
        y1 = kde(x) * scale + bottom
        return x, y1, kde.factor


class Area(LineFillBase[Line]):
    def __init__(
        self,
        x: ArrayLike1D,
        y: ArrayLike1D,
        bottom: ArrayLike1D | float = 0.0,
        orient: OrientationLike = "vertical",
        name: str | None = None,
        backend: Backend | str | None = None,
    ):
        if name is None:
            name = "kde"
        ori = Orientation.parse(orient)
        line = Line(x, y + bottom, name="line", backend=backend)
        fill = Band(x, bottom, y + bottom, name="fill", orient=ori, backend=backend)
        super().__init__(line, fill, name=name)

    @property
    def data(self) -> XYData:
        """The data used to plot the histogram."""
        x, y0, y1 = self.fill.data
        return XYData(x, y1 - y0)

    @data.setter
    def data(self, data: XYData):
        x, y = data
        bottom = self.fill.data.y0
        self.line.data = x, y + bottom
        self.fill.data = x, bottom, y + bottom

    @property
    def orient(self) -> Orientation:
        """Orientation of the line and fill layers."""
        return self.fill.orient



from functools import singledispatch
from typing import TYPE_CHECKING

from plotly import graph_objs as go

from whitecanvas.backend.plotly._base import (
    to_plotly_linestyle,
    to_plotly_marker_symbol,
)
from whitecanvas.layers import _legend as _leg
from whitecanvas.utils.normalize import hex_color

if TYPE_CHECKING:
    from plotly.basedatatypes import BaseTraceType


@singledispatch
def make_sample_item(item) -> "BaseTraceType | None":
    return None


@make_sample_item.register
def _(item: _leg.LineLegendItem):
    return go.Scatter(
        x=[None],
        y=[None],
        mode="lines",
        line={
            "color": hex_color(item.color),
            "width": item.width,
            "dash": to_plotly_linestyle(item.style),
        },
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.MarkersLegendItem):
    return go.Scatter(
        x=[None],
        y=[None],
        mode="markers",
        marker={
            "color": hex_color(item.face.color),
            "size": item.size,
            "symbol": to_plotly_marker_symbol(item.symbol),
            "line": {"color": hex_color(item.edge.color), "width": item.edge.width},
        },
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.ErrorLegendItem):
    # NOTE: plotly errorbars does not support dash
    return go.Scatter(
        x=[None],
        y=[None],
        mode="markers",
        line={"color": hex_color(item.color), "width": item.width},
        marker={"color": "#000000", "size": 0},
        error_y={"type": "data", "array": None, "visible": True},
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.LineErrorLegendItem):
    line = item.line
    if item.xerr is not None:
        error_x = {
            "array": [None],
            "visible": True,
            "color": hex_color(item.xerr.color),
            "width": item.xerr.width,
        }
    else:
        error_x = None
    if item.yerr is not None:
        error_y = {
            "array": [None],
            "visible": True,
            "color": hex_color(item.yerr.color),
            "width": item.yerr.width,
        }
    else:
        error_y = None
    return go.Scatter(
        x=[None],
        y=[None],
        mode="lines",
        line={
            "color": hex_color(line.color),
            "width": line.width,
            "dash": to_plotly_linestyle(line.style),
        },
        error_y=error_y,
        error_x=error_x,
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.MarkerErrorLegendItem):
    marker = item.markers
    if item.xerr is not None:
        error_x = {
            "array": [None],
            "visible": True,
            "color": hex_color(item.xerr.color),
            "width": item.xerr.width,
        }
    else:
        error_x = None
    if item.yerr is not None:
        error_y = {
            "array": [None],
            "visible": True,
            "color": hex_color(item.yerr.color),
            "width": item.yerr.width,
        }
    else:
        error_y = None
    return go.Scatter(
        x=[None],
        y=[None],
        mode="markers",
        marker={
            "color": hex_color(marker.face.color),
            "size": marker.size,
            "symbol": to_plotly_marker_symbol(marker.symbol),
            "line": {"color": hex_color(marker.edge.color), "width": marker.edge.width},
        },
        error_y=error_y,
        error_x=error_x,
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.PlotLegendItem):
    return _make_plot_item(item.line, item.markers)


@make_sample_item.register
def _(item: _leg.PlotErrorLegendItem):
    line = item.plot.line
    marker = item.plot.markers
    if item.xerr is not None:
        error_x = {
            "array": [None],
            "visible": True,
            "color": hex_color(item.xerr.color),
            "width": item.xerr.width,
        }
    else:
        error_x = None
    if item.yerr is not None:
        error_y = {
            "array": [None],
            "visible": True,
            "color": hex_color(item.yerr.color),
            "width": item.yerr.width,
        }
    else:
        error_y = None
    return go.Scatter(
        x=[None],
        y=[None],
        mode="lines+markers",
        line={
            "color": hex_color(line.color),
            "width": line.width,
            "dash": to_plotly_linestyle(line.style),
        },
        marker={
            "color": hex_color(marker.face.color),
            "size": marker.size,
            "symbol": to_plotly_marker_symbol(marker.symbol),
            "line": {"color": hex_color(marker.edge.color), "width": marker.edge.width},
        },
        error_y=error_y,
        error_x=error_x,
        showlegend=True,
    )


def _make_plot_item(line: _leg.LineLegendItem, marker: _leg.MarkersLegendItem):
    return go.Scatter(
        x=[None],
        y=[None],
        mode="lines+markers",
        line={
            "color": hex_color(line.color),
            "width": line.width,
            "dash": to_plotly_linestyle(line.style),
        },
        marker={
            "color": hex_color(marker.face.color),
            "size": marker.size,
            "symbol": to_plotly_marker_symbol(marker.symbol),
            "line": {"color": hex_color(marker.edge.color), "width": marker.edge.width},
        },
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.StemLegendItem):
    line = item.line
    marker = item.markers
    return go.Scatter(
        x=[None],
        y=[None],
        mode="lines+markers",
        line={
            "color": hex_color(line.color),
            "width": line.width,
            "dash": to_plotly_linestyle(line.style),
        },
        marker={
            "color": hex_color(marker.face.color),
            "size": marker.size,
            "symbol": to_plotly_marker_symbol(marker.symbol),
            "line": {"color": hex_color(marker.edge.color), "width": marker.edge.width},
        },
        showlegend=True,
    )


@make_sample_item.register
def _(item: _leg.StemLegendItem):
    return _make_plot_item(item.line, item.markers)


@make_sample_item.register
def _(item: _leg.TitleItem):
    return go.Scatter(
        x=[None],
        y=[None],
        mode="markers",
        marker={"color": "#000000", "size": 0},
        showlegend=True,
    )



from __future__ import annotations

import os
import re
import warnings
from typing import TYPE_CHECKING, Any


warnings.simplefilter("ignore", DeprecationWarning)

if TYPE_CHECKING:
    from mkdocs.structure.pages import Page


def on_page_markdown(md: str, page: Page, **kwargs: Any) -> str:
    """Called when mkdocs is building the markdown for a page."""

    def _add_images(matchobj: re.Match[str]) -> str:
        prefix = matchobj.group(0).split("\n", 1)[0]  # ``` python ...`
        code: str = matchobj.group(1).strip()  # source code

        if code.startswith("#!name:"):
            code, name = _get_image_name(code)
            code, width = _get_image_width(code)

            reldepth = "../" * page.file.src_path.count(os.sep)
            dest = f"{reldepth}_images/{name}.png"
            link = f"\n![]({dest}){{ loading=lazy, width={width}px }}\n\n"
            new_md = f"{prefix}\n{code}\n```{link}"
            return new_md
        elif code.startswith("#!html:"):
            code, name = _get_html_name(code)
            reldepth = "../" * (page.file.src_path.count(os.sep) + 1)
            dest = f"{reldepth}_images/{name}.html"
            html_text = (
                f'<iframe src={dest} frameborder="0" width="400px" height="300px" '
                'scrolling="no"></iframe>'
            )
            new_md = f"{prefix}\n{code}\n```\n\n{html_text}\n"
            return new_md
        elif code.startswith("#!"):
            _, other = code.split("\n", 1)
        else:
            other = code
        return f"{prefix}\n{other}\n```"


    # md = re.sub("``` ?python\n([^`]*)```", _add_images, md, flags=re.DOTALL)
    md = re.sub("``` ?python.*?\n([^`]*)```", _add_images, md)

    return md

def _get_image_name(code: str) -> tuple[str, str]:
    line, other = code.split("\n", 1)
    assert line.startswith("#!name:")
    name = line.split(":", 1)[1].strip()
    return other, name

def _get_html_name(code: str) -> tuple[str, str]:
    line, other = code.split("\n", 1)
    assert line.startswith("#!html:")
    name = line.split(":", 1)[1].strip()
    return other, name

def _get_image_width(code: str) -> tuple[str, int]:
    """Get the width of the image from the code."""
    code = code.strip()
    if code.startswith("#!width:"):
        line, other = code.split("\n", 1)
        width = int(line.split(":", 1)[1].strip())
    else:
        other = code
        width = 360
    return other, width



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any, Sequence, SupportsIndex, TypeVar, overload

from psygnal import Signal, SignalGroup

from whitecanvas import layers as _l
from whitecanvas._axis import DimAxis, RangeAxis
from whitecanvas._exceptions import ReferenceDeletedError
from whitecanvas.layers import _ndim as _ndl
from whitecanvas.types import (
    Alignment,
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    OrientationLike,
    Symbol,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas._base import CanvasBase

_T = TypeVar("_T")


class DimsEvents(SignalGroup):
    """Event signals for `Dims`."""

    indices = Signal(dict)
    axis_names = Signal(list)


_default = object()


class Dims:
    """
    Multi-dimensional plotting interface.
    """

    def __init__(self, canvas: CanvasBase | None = None):
        self._instances: dict[int, Self] = {}
        self._axes: list[DimAxis] = []
        self.events = DimsEvents()
        if canvas is not None:
            self._canvas_ref = weakref.ref(canvas)
            self.events.indices.connect(canvas._draw_canvas, unique=True, max_args=0)
        else:
            self._canvas_ref = lambda: None

    def __get__(self, canvas, owner) -> Self:
        if canvas is None:
            return self
        _id = id(canvas)
        if (ns := self._instances.get(_id)) is None:
            ns = self._instances[_id] = type(self)(canvas)
        return ns

    def _get_canvas(self) -> CanvasBase:
        l = self._canvas_ref()
        if l is None:
            raise ReferenceDeletedError("Canvas has been deleted.")
        return l

    @property
    def names(self) -> list[str]:
        """Names of the dimensions."""
        return [a.name for a in self._axes]

    @overload
    def axis(self, name: str) -> DimAxis: ...

    @overload
    def axis(self, name: str, *, default: _T) -> DimAxis | _T: ...

    def axis(self, name, *, default=_default):
        """Get the axis by name."""
        for a in self._axes:
            if a.name == name:
                return a
        if default is _default:
            raise ValueError(f"Axis {name!r} not found.")
        return default

    @property
    def indices(self) -> dict[str, SupportsIndex]:
        """Current indices for each dimension."""
        return {a.name: a.value() for a in self._axes}

    @overload
    def set_indices(self, arg: dict[str, SupportsIndex]): ...

    @overload
    def set_indices(self, arg: Sequence[SupportsIndex]): ...

    @overload
    def set_indices(self, arg: SupportsIndex): ...

    @overload
    def set_indices(self, **kwargs: SupportsIndex): ...

    def set_indices(self, arg=None, **kwargs: Any):
        """
        Set current indices and update the canvas.

        Either of the following forms are allowed:
        >>> canvas.dims.set_indices(axis_0=3, axis_1=4)
        >>> canvas.dims.set_indices({"axis_0": 3, "axis_1": 4})
        >>> canvas.dims.set_indices([3, 4])
        >>> canvas.dims.set_indices(1)
        """
        if isinstance(arg, dict):
            if kwargs:
                raise TypeError("Cannot specify both positional and keyword arguments.")
            kwargs = arg
        elif isinstance(arg, (list, tuple)):
            names = self.names
            if len(names) < len(arg):
                raise ValueError(
                    f"Number of indices ({len(arg)}) exceeds number of dimensions "
                    f"({len(names)})."
                )
            if kwargs:
                raise TypeError("Cannot specify both positional and keyword arguments.")
            kwargs = dict(zip(names, arg))
        elif arg is None:
            pass
        elif hasattr(arg, "__index__"):
            if kwargs:
                raise TypeError("Cannot specify both positional and keyword arguments.")
            kwargs = {self.names[0]: arg}
        else:
            raise TypeError(
                f"Argument must be dict, list, tuple, or None but got {arg!r}."
            )
        for k, v in kwargs.items():
            self.axis(k).set_value(v)
        self.events.indices.emit(self.indices)

    def in_axes(self, *names: str) -> InAxes:
        """
        Add multi-dimensional layers in the given axis names.

        >>> canvas.dims.in_axes("time").add_line(
        ...     x, [np.sin(x + x0) for x0 in [0, 1, 2]]
        >>> )
        """
        n_names = len(names)
        if n_names == 0:
            raise ValueError("At least one axis name must be provided.")
        if n_names == 1 and isinstance(names[0], (list, tuple)):
            names = names[0]
        existing_names = self.names
        new_axes: list[str] = []
        for name in names:
            if not isinstance(name, str):
                raise TypeError(f"Axis name must be str but got {name!r}.")
            if name not in existing_names:
                new_axes.append(RangeAxis(name, 0))
        self._axes.extend(new_axes)
        return InAxes(self, names)

    def create_widget(self):
        from whitecanvas.backend._window import make_dim_slider

        canvas = self._get_canvas()
        sl = make_dim_slider(canvas, canvas._get_backend().app)
        return sl

    def add_line(
        self,
        xdata: Any,
        ydata: Any,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        alpha: float = 1.0,
        antialias: bool = True,
    ):
        return InAxes(self).add_line(
            xdata, ydata, name=name, color=color, width=width,
            style=style, alpha=alpha, antialias=antialias,
        )  # fmt: skip

    def add_markers(
        self,
        xdata: Any,
        ydata: Any,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        hatch: str | Hatch = Hatch.SOLID,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: float = 5.0,
    ):
        return InAxes(self).add_markers(
            xdata, ydata, name=name, color=color, hatch=hatch,
            symbol=symbol, size=size,
        )  # fmt: skip

    def add_band(
        self,
        xdata: Any,
        ylow: Any,
        yhigh: Any,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ):
        return InAxes(self).add_band(
            xdata,
            ylow,
            yhigh,
            name=name,
            color=color,
            alpha=alpha,
            hatch=hatch,
        )

    def add_errorbars(
        self,
        xdata: Any,
        ylow: Any,
        yhigh: Any,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "blue",
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = False,
        capsize: float = 0.0,
    ):
        return InAxes(self).add_errorbars(
            xdata,
            ylow,
            yhigh,
            name=name,
            orient=orient,
            color=color,
            width=width,
            style=style,
            antialias=antialias,
            capsize=capsize,
        )

    def add_rug(
        self,
        events: Any,
        *,
        low: float = 0.0,
        high: float = 1.0,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "black",
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        alpha: float = 1.0,
    ):
        return InAxes(self).add_rug(
            events, low=low, high=high, name=name, orient=orient, color=color,
            width=width, style=style, antialias=antialias, alpha=alpha
        )  # fmt: skip

    def add_text(
        self,
        xdata: Any,
        ydata: Any,
        string: Any,
        *,
        name: str | None = None,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        family: str | None = None,
    ):
        return InAxes(self).add_text(
            xdata, ydata, string, name=name, color=color, size=size,
            rotation=rotation, anchor=anchor, family=family,
        )  # fmt: skip

    def add_image(
        self,
        image: Any,
        *,
        name: str | None = None,
        cmap: ColormapType = "gray",
        clim: tuple[float | None, float | None] | None = None,
        rgb: bool = False,
        flip_canvas: bool = True,
        lock_aspect: bool = True,
    ):
        return InAxes(self).add_image(
            image, name=name, cmap=cmap, clim=clim, rgb=rgb,
            flip_canvas=flip_canvas, lock_aspect=lock_aspect,
        )  # fmt: skip


class InAxes:
    def __init__(self, dims: Dims, axis_names: list[str] | None = None):
        self._dims = dims
        self._axis_names = axis_names

    def _get_canvas(self) -> CanvasBase:
        return self._dims._get_canvas()

    def add_line(
        self,
        xdata: Any,
        ydata: Any,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        alpha: float = 1.0,
        antialias: bool = True,
    ):
        """
        Add multi-dimensional line layer.

        Parameters
        ----------
        xdata : Any
            X data of multi-dimensional structure.
        ydata : Any
            Y data of multi-dimensional structure.

        Returns
        -------
        XYLayerStack
            Layer stack of a line layer.
        """
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        color = canvas._generate_colors(color)
        stack = _ndl.XYLayerStack.from_layer_class(
            _l.Line, xdata, ydata, name=name, color=color, width=width, style=style,
            alpha=alpha, antialias=antialias, backend=canvas._get_backend(),
        )  # fmt: skip
        return self._add_layer(stack)

    def add_markers(
        self,
        xdata: Any,
        ydata: Any,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        hatch: str | Hatch = Hatch.SOLID,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: float = 5.0,
    ) -> _ndl.XYLayerStack:
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        color = canvas._generate_colors(color)
        stack = _ndl.XYLayerStack.from_layer_class(
            _l.Markers, xdata, ydata, name=name, color=color, hatch=hatch,
            symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        return self._add_layer(stack)

    def add_band(
        self,
        xdata: Any,
        ylow: Any,
        yhigh: Any,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _ndl.XYYLayerStack:
        """
        Add multi-dimensional band layer.

        Parameters
        ----------
        xdata : Any
            X data of multi-dimensional structure.
        ylow : Any
            Lower y data of multi-dimensional structure.
        yhigh : Any
            Higher y data of multi-dimensional structure.

        Returns
        -------
        XYYLayerStack
            Layer stack of a band layer.
        """
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        color = canvas._generate_colors(color)
        stack = _ndl.XYYLayerStack.from_layer_class(
            _l.Band, xdata, ylow, yhigh, name=name, color=color, hatch=hatch,
            alpha=alpha, backend=canvas._get_backend(),
        )  # fmt: skip
        return self._add_layer(stack)

    def add_errorbars(
        self,
        xdata: Any,
        ylow: Any,
        yhigh: Any,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "blue",
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = False,
        capsize: float = 0.0,
    ) -> _ndl.XYYLayerStack:
        """
        Add multi-dimensional errorbars layer.

        Parameters
        ----------
        xdata : Any
            X data of multi-dimensional structure.
        ylow : Any
            Lower y data of multi-dimensional structure.
        yhigh : Any
            Higher y data of multi-dimensional structure.

        Returns
        -------
        XYYLayerStack
            Layer stack of an errorbars layer.
        """
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        color = canvas._generate_colors(color)
        stack = _ndl.XYYLayerStack.from_layer_class(
            _l.Errorbars, xdata, ylow, yhigh, name=name, orient=orient, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            backend=canvas._get_backend(),
        )  # fmt: skip
        return self._add_layer(stack)

    def add_rug(
        self,
        events: Any,
        *,
        low: float = 0.0,
        high: float = 1.0,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "black",
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        alpha: float = 1.0,
    ) -> _ndl.YLayerStack:
        """
        Add multi-dimensional rug layer.

        Parameters
        ----------
        events : Any
            Multi-dimensional event data.

        Returns
        -------
        YLayerStack
            Layer stack of a rug layer.
        """
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        color = canvas._generate_colors(color)
        stack = _ndl.YLayerStack.from_layer_class(
            _l.Rug, events, low=low, high=high, name=name, orient=orient,
            color=color,  width=width,  style=style, antialias=antialias,
            alpha=alpha, backend=canvas._get_backend(),
        )  # fmt: skip
        return self._add_layer(stack)

    def add_text(
        self,
        xdata: Any,
        ydata: Any,
        string: Any,
        *,
        name: str | None = None,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        family: str | None = None,
    ) -> _ndl.TextLayerStack:
        """
        Add multi-dimensional text layer.

        Parameters
        ----------
        xdata : Any
            _description_
        ydata : Any
            _description_
        string : Any
            _description_

        Returns
        -------
        TextLayerStack
            Layer stack of text layers.
        """
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        stack = _ndl.TextLayerStack.from_layer_class(
            _l.Texts, xdata, ydata, string, name=name, color=color, size=size,
            rotation=rotation, anchor=anchor, family=family,
            backend=canvas._get_backend(),
        )  # fmt: skip
        return self._add_layer(stack)

    def add_image(
        self,
        image: Any,
        *,
        name: str | None = None,
        cmap: ColormapType = "gray",
        clim: tuple[float | None, float | None] | None = None,
        rgb: bool = False,
        flip_canvas: bool = True,
        lock_aspect: bool = True,
    ) -> _ndl.ImageLayerStack:
        """
        Add multi-dimensional image layer.

        Parameters
        ----------
        image : Any
            Image stack of multi-dimensional structure.
        rgb : bool, default False
            If input image is an RGB(A) image of shape (..., 3) or (..., 4),
            then set this to True to display the image as RGB(A).

        Returns
        -------
        ImageLayerStack
            Layer stack of image layers.
        """
        canvas = self._get_canvas()
        name = canvas._coerce_name(name)
        stack = _ndl.ImageLayerStack.from_layer_class(
            _l.Image, image, name=name, cmap=cmap, clim=clim, rgb=rgb,
            backend=canvas._get_backend(),
        )  # fmt: skip
        self._add_layer(stack)

        if flip_canvas and not canvas.y.flipped:
            canvas.y.flipped = True
        if lock_aspect:
            canvas.aspect_ratio = 1.0
        return stack

    def _add_layer(self, stack: _l.LayerStack):
        canvas = self._get_canvas()
        if self._axis_names is not None:
            stack._with_axis_names(self._axis_names)
        new_axes = []
        for a, s in zip(stack.axis_names, stack._data_stack.shape):
            axis = self._dims.axis(a, default=None)
            if axis is None:
                new_axes.append(RangeAxis(a, s))
            elif axis.value() < s:
                if isinstance(axis, RangeAxis):
                    axis.set_size(s)
                else:
                    raise NotImplementedError
        self._dims._axes.extend(new_axes)
        if len(new_axes) > 0:
            self._dims.events.axis_names.emit(self._dims.names)
        return canvas.add_layer(stack)


for meth in dir(InAxes):
    if meth.startswith("add_"):
        doc = getattr(InAxes, meth).__doc__
        getattr(Dims, meth).__doc__ = doc



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from plotly import graph_objects as go

from whitecanvas.backend import _not_implemented
from whitecanvas.backend.plotly._base import PlotlyHoverableLayer
from whitecanvas.protocols import BarProtocol, check_protocol
from whitecanvas.types import Hatch
from whitecanvas.utils.normalize import arr_color, rgba_str_color


@check_protocol(BarProtocol)
class Bars(PlotlyHoverableLayer[go.Bar]):
    def __init__(self, xlow, xhigh, ylow, yhigh):
        x = (xlow + xhigh) / 2
        ndata = len(x)
        self._props = {
            "x": x,
            "y": yhigh - ylow,
            "width": xhigh - xlow,
            "base": ylow,
            "marker": {
                "color": ["blue"] * ndata,
                "pattern": {
                    "shape": "-",
                },
                "line": {
                    "width": [1] * ndata,
                    "color": ["blue"] * ndata,
                },
            },
            "visible": True,
            "showlegend": False,
            "customdata": [""] * ndata,
            "hovertemplate": "%{customdata}<extra></extra>",
            "type": "bar",
        }
        PlotlyHoverableLayer.__init__(self)

    ##### XXYYDataProtocol #####
    def _plt_get_data(self):
        half_width = self._props["width"] / 2
        x = self._props["x"]
        yhigh = self._props["y"]
        # for some reason, the base property is dtype=object
        ylow = self._props["base"].astype(yhigh.dtype, copy=False)
        xlow = x - half_width
        xhigh = x + half_width
        return xlow, xhigh, ylow, yhigh

    def _plt_set_data(self, x0, x1, y0, y1):
        self._props["x"] = (x0 + x1) / 2
        self._props["y"] = y1 - y0
        self._props["width"] = x1 - x0
        self._props["base"] = y0

    def _plt_get_ndata(self):
        return len(self._props["x"])

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        color = self._props["marker"]["color"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            color = [rgba_str_color(color)] * self._plt_get_ndata()
        else:
            color = [rgba_str_color(c) for c in color]
        self._props["marker"]["color"] = color

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [Hatch(p) for p in self._props["marker"]["pattern"]["shape"]]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if isinstance(pattern, Hatch):
            ptn = [pattern.value] * self._plt_get_ndata()
        else:
            ptn = [ptn.value for ptn in pattern]
        self._props["marker"]["pattern"]["shape"] = ptn

    ##### HasEdges protocol #####

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return np.array(self._props["marker"]["line"]["width"])

    def _plt_set_edge_width(self, width: float):
        if np.isscalar(width):
            width = np.full(self._plt_get_ndata(), width)
        self._props["marker"]["line"]["width"] = width

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_styles()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        color = self._props["marker"]["line"]["color"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            color = [rgba_str_color(color)] * self._plt_get_ndata()
        else:
            color = [rgba_str_color(c) for c in color]
        self._props["marker"]["line"]["color"] = color



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Iterator, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas.utils.type_check import is_pandas_dataframe, is_polars_dataframe

if TYPE_CHECKING:
    import pandas as pd  # noqa: F401
    import polars as pl  # noqa: F401
    from typing_extensions import Self

_T = TypeVar("_T")


class DataFrameWrapper(ABC, Generic[_T]):
    def __init__(self, data: _T):
        self._data = data

    def __repr__(self) -> str:
        return f"{type(self).__name__} of {self._data!r}"

    def __len__(self) -> int:
        such_as = next(self.iter_values(), None)
        if such_as is None:
            return 0
        else:
            return such_as.size

    @property
    def shape(self) -> tuple[int, int]:
        such_as = next(self.iter_values(), None)
        if such_as is None:
            return 0, 0
        else:
            return such_as.size, len(self.columns)

    def get_native(self) -> _T:
        return self._data

    @abstractmethod
    def __getitem__(self, item: str) -> NDArray[np.generic]:
        ...

    def __contains__(self, item: str) -> bool:
        return item in self.iter_keys()

    @abstractmethod
    def iter_keys(self) -> Iterator[str]:
        ...

    def iter_values(self) -> Iterator[NDArray[np.generic]]:
        for k in self.iter_keys():
            yield self[k]

    def iter_items(self) -> Iterator[tuple[str, NDArray[np.generic]]]:
        for k in self.iter_keys():
            yield k, self[k]

    @abstractmethod
    def select(self, columns: list[str]) -> Self:
        """Select columns from the data frame and return a new one."""

    @abstractmethod
    def sort(self, by: str) -> Self:
        """Sort the data frame by a column and return the sorted one."""

    @abstractmethod
    def get_rows(self, indices: list[int]) -> Self:
        """Get rows by indices and return a new data frame."""

    @abstractmethod
    def filter(
        self,
        by: tuple[str, ...],
        values: tuple[Any, ...],
    ) -> Self:
        ...

    @abstractmethod
    def group_by(self, by: tuple[str, ...]) -> Iterator[tuple[tuple[Any, ...], Self]]:
        ...

    @abstractmethod
    def agg_by(self, by: tuple[str, ...], on: list[str], method: str) -> Self:
        ...

    @abstractmethod
    def melt(self, id_vars: list[str], value_vars: list[str]) -> Self:
        ...

    @abstractmethod
    def value_count(self, by: tuple[str, ...]) -> Self:
        """Return the count of each group."""

    @abstractmethod
    def value_first(self, by: tuple[str, ...], on: str) -> Self:
        """Return the first value of a column for each group."""

    @property
    def columns(self) -> list[str]:
        """Column names of the data frame."""
        return list(self.iter_keys())


class DictWrapper(DataFrameWrapper[dict[str, np.ndarray]]):
    def __getitem__(self, item: str) -> np.ndarray:
        if not isinstance(item, str):
            raise TypeError(f"Unsupported type: {type(item)}")
        try:
            return self._data[item]
        except KeyError:
            raise KeyError(
                f"{item!r} not in the keys. Valid keys are: {list(self._data.keys())}."
            ) from None

    def iter_keys(self) -> Iterator[str]:
        return iter(self._data.keys())

    def select(self, columns: list[str]) -> Self:
        return DictWrapper({k: v for k, v in self._data.items() if k in columns})

    def sort(self, by: str) -> Self:
        arr = self[by]
        indices = np.argsort(arr)
        return DictWrapper({k: v[indices] for k, v in self._data.items()})

    def get_rows(self, indices: list[int]) -> Self:
        return DictWrapper({k: v[indices] for k, v in self._data.items()})

    def filter(
        self,
        by: tuple[str, ...],
        values: tuple[Any, ...],
    ) -> Self:
        sers = np.column_stack([self._data[b] == val for b, val in zip(by, values)])
        sl = sers.all(axis=1)
        return DictWrapper({k: v[sl] for k, v in self._data.items()})

    def group_by(self, by: tuple[str, ...]) -> Iterator[tuple[tuple[Any, ...], Self]]:
        if by == ():
            yield (), self
            return
        observed = set()
        for row in zip(*[self._data[b] for b in by]):
            if row in observed:
                continue
            yield row, self.filter(by, row)
            observed.add(row)

    def agg_by(self, by: tuple[str, ...], on: list[str], method: str) -> Self:
        if method not in ("min", "max", "mean", "median", "sum", "std"):
            raise ValueError(f"Unsupported aggregation method: {method}")
        agg = getattr(np, method)
        out = {k: [] for k in (*by, *on)}
        for sl, sub in self.group_by(by):
            for b, s in zip(by, sl):
                out[b].append(s)
            for o in on:
                out[o].append(agg(sub[o]))
        return DictWrapper({k: np.array(v) for k, v in out.items()})

    def melt(
        self,
        id_vars: list[str],
        value_vars: list[str],
        var_name: str | None = None,
        value_name: str | None = None,
    ) -> Self:
        if var_name is None:
            var_name = "variable"
        if value_name is None:
            value_name = "value"
        out = {k: [] for k in [*id_vars, var_name, value_name]}
        for k, v in self._data.items():
            if k in id_vars:
                out[k].extend(v)
            elif k in value_vars:
                out[var_name] += [k] * len(v)
                out[value_name].extend(v)
            else:
                pass
        return DictWrapper({k: np.array(v) for k, v in out.items()})

    def value_count(self, by: tuple[str, ...]) -> Self:
        out = {k: [] for k in [*by, "size"]}
        for sl, sub in self.group_by(by):
            for b, s in zip(by, sl):
                out[b].append(s)
            out["size"].append(len(sub[by[0]]))
        return DictWrapper({k: np.array(v) for k, v in out.items()})

    def value_first(self, by: tuple[str, ...], on: str) -> Self:
        out = {k: [] for k in [*by, on]}
        for sl, sub in self.group_by(by):
            for b, s in zip(by, sl):
                out[b].append(s)
            out[on].append(sub[on][0])
        return DictWrapper({k: np.array(v) for k, v in out.items()})


class PandasWrapper(DataFrameWrapper["pd.DataFrame"]):
    def __getitem__(self, item: str) -> np.ndarray:
        if not isinstance(item, str):
            raise TypeError(f"Unsupported type: {type(item)}")
        series = self._data[item]
        if series.size > 0 and isinstance(series.iloc[0], str):
            return series.to_numpy().astype(str)
        else:
            return series.to_numpy()

    def iter_keys(self) -> Iterator[str]:
        return iter(self._data.columns)

    def select(self, columns: list[str]) -> Self:
        return PandasWrapper(self._data[columns])

    def sort(self, by: str) -> Self:
        return PandasWrapper(self._data.sort_values(by))

    def get_rows(self, indices: list[int]) -> Self:
        return PandasWrapper(self._data.iloc[indices])

    def filter(
        self,
        by: tuple[str, ...],
        values: tuple[Any, ...],
    ) -> Self:
        sers = np.column_stack([self._data[b] == val for b, val in zip(by, values)])
        return PandasWrapper(self._data[sers])

    def group_by(self, by: tuple[str, ...]) -> Iterator[tuple[tuple[Any, ...], Self]]:
        if by == ():
            yield (), self
            return
        for sl, sub in self._data.groupby(list(by), observed=True):
            yield sl, PandasWrapper(sub)

    def agg_by(self, by: tuple[str, ...], on: list[str], method: str) -> Self:
        on = list(on)
        return PandasWrapper(self._data.groupby(list(by))[on].agg(method).reset_index())

    def melt(
        self,
        id_vars: list[str],
        value_vars: list[str],
        var_name: str | None = None,
        value_name: str | None = None,
    ) -> Self:
        if value_name is None:
            value_name = "value"
        return PandasWrapper(
            self._data.melt(
                id_vars=id_vars,
                value_vars=value_vars,
                var_name=var_name,
                value_name=value_name,
                ignore_index=True,
            )
        )

    def value_count(self, by: tuple[str, ...]) -> Self:
        import pandas as pd  # noqa: F811, RUF100

        rows = []
        for sl, sub in self._data.groupby(list(by), observed=True):
            rows.append((*sl, len(sub)))
        return PandasWrapper(pd.DataFrame(rows, columns=[*by, "size"]))

    def value_first(self, by: tuple[str, ...], on: str) -> Self:
        return PandasWrapper(self._data.groupby(list(by)).first().reset_index())


class PolarsWrapper(DataFrameWrapper["pl.DataFrame"]):
    def __getitem__(self, item: str) -> np.ndarray:
        if not isinstance(item, str):
            raise TypeError(f"Unsupported type: {type(item)}")
        try:
            return self._data[item].to_numpy()
        except Exception as e:
            import polars as pl  # noqa: F811, RUF100

            if isinstance(e, pl.ColumnNotFoundError):
                raise KeyError(item) from None
            raise e

    def iter_keys(self) -> Iterator[str]:
        return iter(self._data.columns)

    def select(self, columns: list[str]) -> Self:
        return PolarsWrapper(self._data.select(columns))

    def sort(self, by: str) -> Self:
        return PolarsWrapper(self._data.sort(by))

    def get_rows(self, indices: list[int]) -> Self:
        return PolarsWrapper(self._data[indices])

    def filter(
        self,
        by: tuple[str, ...],
        values: tuple[Any, ...],
    ) -> Self:
        kwargs = dict(zip(by, values))
        df = self._data.filter(**kwargs)
        return PolarsWrapper(df)

    def group_by(self, by: tuple[str, ...]) -> Iterator[tuple[tuple[Any, ...], Self]]:
        if by == ():
            yield (), self
            return
        for sl, sub in self._data.group_by(by, maintain_order=True):
            yield sl, PolarsWrapper(sub)

    def agg_by(self, by: tuple[str, ...], on: list[str], method: str) -> Self:
        import polars as pl  # noqa: F811, RUF100

        exprs = [getattr(pl.col(o), method)() for o in on]
        return PolarsWrapper(self._data.group_by(by, maintain_order=True).agg(*exprs))

    def melt(
        self,
        id_vars: list[str],
        value_vars: list[str],
        var_name: str | None = None,
        value_name: str | None = None,
    ) -> Self:
        return PolarsWrapper(
            self._data.melt(
                id_vars=id_vars,
                value_vars=value_vars,
                var_name=var_name,
                value_name=value_name,
            )
        )

    def value_count(self, by: tuple[str, ...]) -> Self:
        return PolarsWrapper(
            self._data.group_by(by, maintain_order=True)
            .count()
            .rename({"count": "size"})
        )

    def value_first(self, by: tuple[str, ...], on: str) -> Self:
        return PolarsWrapper(self._data.group_by(by, maintain_order=True).first())


def parse(data: Any) -> DataFrameWrapper:
    """Parse a data object into a DataFrameWrapper."""
    if isinstance(data, DataFrameWrapper):
        return data
    if isinstance(data, dict):
        df = {k: np.asarray(v) for k, v in data.items()}
        if len(df) > 0:  # check length
            if len({v.size for v in df.values()}) > 1:
                raise ValueError("All columns must have the same length")
        return DictWrapper(df)
    elif is_pandas_dataframe(data):
        return PandasWrapper(data)
    elif is_polars_dataframe(data):
        return PolarsWrapper(data)
    else:
        raise TypeError(f"Unsupported data type: {type(data)}")



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

import numpy as np
from vispy import scene
from vispy.visuals.axis import AxisVisual, Ticker

from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    from vispy.visuals import LineVisual, TextVisual

    from whitecanvas.backend.vispy.canvas import Camera, Canvas


class TextLabel(scene.Label):
    _text_visual: TextVisual

    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    def _plt_get_text(self) -> str:
        return self.text

    def _plt_set_text(self, text: str):
        self.text = text

    def _plt_get_color(self):
        return np.array(self._text_visual.color).ravel()

    def _plt_set_color(self, color):
        self._text_visual.color = color

    def _plt_get_size(self) -> int:
        return self._text_visual.font_size

    def _plt_set_size(self, size: int):
        self._text_visual.font_size = size

    def _plt_get_fontfamily(self) -> str:
        return self._text_visual.face

    def _plt_set_fontfamily(self, family: str):
        self._text_visual.face = family


class Axis(scene.AxisWidget):
    axis: AxisVisual

    def __init__(self, canvas: Canvas, dim: int, **kwargs):
        kwargs.setdefault("axis_width", 2)
        kwargs.setdefault("tick_width", 1)
        super().__init__(**kwargs)
        self.unfreeze()
        self._dim = dim
        self._canvas_ref = weakref.ref(canvas)
        self.freeze()

    def _plt_viewbox(self) -> scene.ViewBox:
        return self._canvas_ref()._viewbox

    def _plt_camera(self) -> Camera:
        return self._plt_viewbox().camera

    def _plt_get_limits(self) -> tuple[float, float]:
        rect = self._plt_camera().rect
        if self._dim == 0:  # y
            return rect.bottom, rect.top
        else:
            return rect.left, rect.right

    def _plt_set_limits(self, limits: tuple[float, float]):
        camera = self._plt_camera()
        # NOTE: margin = 0 is ignored in the current implementation of vispy.
        # use a very small margin instead.
        margin = (limits[1] - limits[0]) * 1e-8
        with camera.changed.blocked():
            if self._dim == 0:  # y
                xlim = self._canvas_ref()._xaxis._plt_get_limits()
                camera.set_range(x=xlim, y=limits, margin=margin)
            else:
                ylim = self._canvas_ref()._yaxis._plt_get_limits()
                camera.set_range(x=limits, y=ylim, margin=margin)
        camera.set_default_state()
        camera.reset()

    def _plt_get_color(self):
        return np.array(self.axis.axis_color).ravel()

    def _plt_set_color(self, color):
        self.axis.axis_color = color

    def _plt_flip(self) -> None:
        camera = self._plt_camera()
        flipped = list(camera.flip)
        axis = 1 - self._dim
        flipped[axis] = not flipped[axis]
        camera.flip = tuple(flipped)

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        grid_lines = self._canvas_ref()._grid_lines
        if self._dim == 0:  # y
            grid_lines.set_y_grid_lines(visible, color, width, style)
        else:
            grid_lines.set_x_grid_lines(visible, color, width, style)


class Ticks:
    def __init__(self, axis: Axis):
        self._axis = weakref.ref(axis)
        axis.axis.ticker = VispyTicker(axis.axis)

    def _get_ticks(self) -> LineVisual:
        return self._axis().axis._ticks

    def _get_ticker(self) -> VispyTicker:
        return self._axis().axis.ticker

    @property
    def _text(self) -> TextVisual:
        return self._axis().axis._text

    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        pos = self._get_ticks().pos
        if pos is None:
            return [], []
        return list(pos), self._text.text

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        self._get_ticker()._categorical_labels = (pos, labels)

    def _plt_reset_override(self):
        self._get_ticker()._categorical_labels = None

    def _plt_get_color(self):
        return np.array(self._text.color)

    def _plt_set_color(self, color):
        self._text.color = color
        self._axis().axis.tick_color = color

    def _plt_get_visible(self) -> bool:
        return self._get_ticker().visible

    def _plt_set_visible(self, visible: bool):
        self._get_ticker().visible = visible
        # axis = self._axis()
        # if axis._dim == 0:  # y
        #     axis.width_min = axis.width_max = 40 if visible else 0
        # else:
        #     axis.height_min = axis.height_max = 50 if visible else 0

    def _plt_get_size(self) -> float:
        return self._text.font_size

    def _plt_set_size(self, size: float):
        self._text.font_size = size

    def _plt_get_fontfamily(self) -> str:
        return self._text.face

    def _plt_set_fontfamily(self, font):
        self._text.face = font

    def _plt_get_text_rotation(self) -> float:
        return -self._text.rotation

    def _plt_set_text_rotation(self, rotation: float):
        self._text.rotation = -rotation


class VispyTicker(Ticker):
    """Ticker that supports categorical axes"""

    axis: AxisVisual

    def __init__(self, axis: AxisVisual):
        anchors = axis.ticker._anchors
        super().__init__(axis, anchors)
        self._categorical_labels: tuple[list[float], list[str]] | None = None
        self._visible = True

    def _get_tick_frac_labels(self):
        if not self._visible:
            major, minor, labels = np.zeros(0), np.zeros(0), np.zeros(0)
        elif self._categorical_labels is None:
            major, minor, labels = super()._get_tick_frac_labels()
        else:
            pos, labels = self._categorical_labels
            domain = self.axis.domain
            scale = domain[1] - domain[0]
            major_tick_fractions = (np.asarray(pos) - domain[0]) / scale
            minor_tick_fractions = np.zeros(0)
            ok = (0 <= major_tick_fractions) & (major_tick_fractions <= 1)
            tick_labels = np.asarray(labels)[ok]
            major = major_tick_fractions[ok]
            minor = minor_tick_fractions
            labels = tick_labels
        return major, minor, labels

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, visible: bool):
        self._visible = visible
        self.axis.update()
        self.axis._update_subvisuals()



from __future__ import annotations

from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray

from whitecanvas.types import HistBinType


class HistogramTuple(NamedTuple):
    edges: NDArray[np.number]
    counts: list[NDArray[np.integer]]

    def density(self) -> list[NDArray[np.number]]:
        return self.probability_scaled(self.widths())

    def probability(self) -> list[NDArray[np.number]]:
        return self.probability_scaled(1)

    def percent(self) -> list[NDArray[np.number]]:
        return self.probability_scaled(100)

    def frequency(self) -> list[NDArray[np.number]]:
        return self.scaled(self.widths())

    def scaled(self, scale: float | NDArray[np.number]) -> list[NDArray[np.number]]:
        out: list[NDArray[np.number]] = []
        for arr in self.counts:
            scaled = arr / scale
            out.append(scaled)
        return out

    def probability_scaled(
        self,
        scale: float | NDArray[np.number],
    ) -> list[NDArray[np.number]]:
        out: list[NDArray[np.number]] = []
        for arr in self.counts:
            density_scaled = arr / arr.sum() / scale
            out.append(density_scaled)
        return out

    def centers(self) -> NDArray[np.number]:
        return (self.edges[:-1] + self.edges[1:]) / 2

    def widths(self) -> NDArray[np.number]:
        return np.diff(self.edges)


def get_hist_edges(
    arrays: list[NDArray[np.number]],
    bins: HistBinType,
    range: tuple[float, float] | None = None,
) -> NDArray[np.number]:
    return np.histogram_bin_edges(np.concatenate(arrays), bins, range)


def histograms(
    arrays: list[NDArray[np.number]],
    bins: HistBinType,
    range: tuple[float, float] | None = None,
) -> HistogramTuple:
    edges = get_hist_edges(arrays, bins, range)
    counts = [np.histogram(arr, edges)[0] for arr in arrays]
    return HistogramTuple(edges, counts)



from whitecanvas.layers._primitive.band import Band
from whitecanvas.layers._primitive.bars import Bars
from whitecanvas.layers._primitive.errorbars import Errorbars
from whitecanvas.layers._primitive.image import Image
from whitecanvas.layers._primitive.inf_curve import InfCurve, InfLine
from whitecanvas.layers._primitive.line import Line, LineStep, MultiLine
from whitecanvas.layers._primitive.markers import Markers
from whitecanvas.layers._primitive.rects import Rects
from whitecanvas.layers._primitive.rug import Rug
from whitecanvas.layers._primitive.spans import Spans
from whitecanvas.layers._primitive.text import Texts
from whitecanvas.layers._primitive.vectors import Vectors

__all__ = [
    "Line",
    "LineStep",
    "MultiLine",
    "Bars",
    "Markers",
    "Band",
    "Spans",
    "Errorbars",
    "Rug",
    "InfLine",
    "InfCurve",
    "Texts",
    "Image",
    "Rects",
    "Vectors",
]



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from plotly import graph_objects as go

from whitecanvas.backend.plotly._base import PlotlyLayer
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, Hatch, LineStyle
from whitecanvas.utils.normalize import arr_color, as_color_array, rgba_str_color


@check_protocol(TextProtocol)
class Texts(PlotlyLayer[go.Scatter]):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        ntexts = len(text)
        self._props = {
            "x": x,
            "y": y,
            "mode": "text",
            "text": text,
            "textposition": "bottom left",
            "textfont": {
                "family": "Arial",
                "size": np.full(ntexts, 10),
                "color": ["rgba(0, 0, 0, 255)"] * ntexts,
            },
            "type": "scatter",
            "showlegend": False,
            # "angle": np.zeros(ntexts),
            "visible": True,
        }
        self._angle = np.zeros(ntexts)

    ##### TextProtocol #####

    def _plt_get_text(self) -> list[str]:
        return list(self._props["text"])

    def _plt_set_text(self, text: list[str]):
        self._props["text"] = text

    def _plt_get_text_color(self):
        return np.stack(
            [arr_color(c) for c in self._props["textfont"]["color"]],
            axis=0,
        )

    def _plt_set_text_color(self, color):
        color = as_color_array(color, len(self._props["text"]))
        self._props["textfont"]["color"] = [rgba_str_color(c) for c in color]

    def _plt_get_text_size(self) -> NDArray[np.floating]:
        return self._props["textfont"]["size"]

    def _plt_set_text_size(self, size: NDArray[np.floating]):
        if np.isscalar(size):
            size = np.full(len(self._props["text"]), size)
        self._props["textfont"]["size"] = size

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        return self._props["x"], self._props["y"]

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        self._props["x"], self._props["y"] = position

    def _plt_get_text_anchor(self) -> Alignment:
        return _from_plotly_alignment(self._props["textposition"])

    def _plt_set_text_anchor(self, anc: Alignment):
        self._props["textposition"] = _to_plotly_alignment(anc)

    def _plt_get_text_rotation(self):
        return self._angle

    def _plt_set_text_rotation(self, rotation: float):
        if np.isscalar(rotation):
            rotation = np.full(len(self._props["text"]), rotation)
        self._angle = rotation

    def _plt_get_text_fontfamily(self) -> str:
        return self._props["textfont"]["family"]

    def _plt_set_text_fontfamily(self, fontfamily: str):
        self._props["textfont"]["family"] = fontfamily

    ##### HasFaces #####

    def _plt_get_face_color(self):
        return np.zeros((len(self._props["text"]), 4))

    def _plt_set_face_color(self, color):
        pass

    def _plt_get_face_hatch(self) -> Hatch:
        return [Hatch.SOLID] * len(self._props["text"])

    def _plt_set_face_hatch(self, pattern: Hatch):
        pass

    ##### HasEdges #####

    def _plt_get_edge_color(self):
        return np.zeros((len(self._props["text"]), 4))

    def _plt_set_edge_color(self, color):
        pass

    def _plt_get_edge_width(self) -> float:
        return np.full(len(self._props["text"]), 0.0)

    def _plt_set_edge_width(self, width: float):
        pass

    def _plt_get_edge_style(self) -> LineStyle:
        return [LineStyle.SOLID] * len(self._props["text"])

    def _plt_set_edge_style(self, style: LineStyle):
        pass


def _from_plotly_alignment(s: str) -> Alignment:
    return _ALIGNMENT_MAP.get(s.lower(), Alignment.CENTER)


def _to_plotly_alignment(a: Alignment) -> str:
    return _ALIGNMENT_MAP_INV.get(a, "center")


_ALIGNMENT_MAP = {
    "top center": Alignment.BOTTOM,
    "top left": Alignment.BOTTOM_RIGHT,
    "top right": Alignment.BOTTOM_LEFT,
    "middle center": Alignment.CENTER,
    "middle left": Alignment.RIGHT,
    "middle right": Alignment.LEFT,
    "bottom center": Alignment.TOP,
    "bottom left": Alignment.TOP_RIGHT,
    "bottom right": Alignment.TOP_LEFT,
}

_ALIGNMENT_MAP_INV = {v: k for k, v in _ALIGNMENT_MAP.items()}



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from plotly import graph_objects as go

from whitecanvas.backend import _not_implemented
from whitecanvas.backend.plotly._base import (
    PlotlyHoverableLayer,
    from_plotly_marker_symbol,
    to_plotly_marker_symbol,
)
from whitecanvas.types import Symbol
from whitecanvas.utils.normalize import arr_color, as_color_array, rgba_str_color
from whitecanvas.utils.type_check import is_real_number


class Markers3D(PlotlyHoverableLayer[go.Scatter3d]):
    def __init__(self, xdata, ydata, zdata):
        ndata = len(xdata)
        self._props = {
            "x": xdata,
            "y": ydata,
            "z": zdata,
            "mode": "markers",
            "marker": {
                "color": ["blue"] * ndata,
                "size": np.full(ndata, 10),
                "symbol": "circle",
                "line": {"width": 1, "color": ["blue"] * ndata},
            },
            "type": "scatter3d",
            "showlegend": False,
            "visible": True,
            "customdata": [""] * ndata,
            "hovertemplate": "%{customdata}<extra></extra>",
        }
        PlotlyHoverableLayer.__init__(self)

    def _plt_get_ndata(self) -> int:
        return len(self._props["x"])

    def _plt_get_data(self):
        return self._props["x"], self._props["y"], self._props["z"]

    def _plt_set_data(self, xdata, ydata, zdata):
        self._props["x"] = xdata
        self._props["y"] = ydata
        self._props["z"] = zdata

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        color = self._props["marker"]["color"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._plt_get_ndata())
        self._props["marker"]["color"] = [rgba_str_color(c) for c in color]

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatches()

    def _plt_get_symbol(self) -> Symbol:
        return from_plotly_marker_symbol(self._props["marker"]["symbol"])

    def _plt_set_symbol(self, symbol: Symbol):
        self._props["marker"]["symbol"] = to_plotly_marker_symbol(symbol)

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return np.asarray(self._props["marker"]["size"])

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(self._plt_get_ndata(), size)
        self._props["marker"]["size"] = size

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return np.full(self._plt_get_ndata, self._props["marker"]["line"]["width"])

    def _plt_set_edge_width(self, width: float):
        if is_real_number(width):
            self._props["marker"]["line"]["width"] = width
        else:
            self._props["marker"]["line"]["width"] = width[0]

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_styles()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        color = self._props["marker"]["line"]["color"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._plt_get_ndata())
        self._props["marker"]["line"]["color"] = [rgba_str_color(c) for c in color]



from __future__ import annotations

import numpy as np
from matplotlib.container import BarContainer
from matplotlib.patches import Rectangle
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import MplMouseEventsMixin
from whitecanvas.protocols import BarProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_not_array


def _rectangle(x, y, dx, dy) -> Rectangle:
    r = Rectangle(xy=(x, y), width=dx, height=dy, linestyle="-", picker=True)
    r.get_path()._interpolation_steps = 100
    return r


@check_protocol(BarProtocol)
class Bars(BarContainer, MplMouseEventsMixin):
    def __init__(self, xlow, xhigh, ylow, yhigh):
        patches = []
        width = xhigh - xlow
        height = yhigh - ylow
        for x, y, dx, dy in zip(xlow, ylow, width, height):
            patches.append(_rectangle(x, y, dx, dy))
        super().__init__(patches)
        self._visible = True
        MplMouseEventsMixin.__init__(self)

    def _plt_get_visible(self):
        return self._visible

    def __iter__(self):
        return iter(self.patches)

    def _plt_set_visible(self, visible):
        for patch in self.patches:
            patch.set_visible(visible)
        self._visible = visible

    def _plt_set_zorder(self, zorder: int):
        for patch in self.patches:
            patch.set_zorder(zorder)

    def get_zorder(self):
        if len(self.patches) == 0:
            return 0
        return self.patches[0].get_zorder()

    ##### XXYYDataProtocol #####
    def _plt_get_data(self):
        n = len(self.patches)
        x0 = np.empty(n, dtype=np.float32)
        x1 = np.empty(n, dtype=np.float32)
        y0 = np.empty(n, dtype=np.float32)
        y1 = np.empty(n, dtype=np.float32)
        for i, patch in enumerate(self.patches):
            x0[i] = patch.get_x()
            x1[i] = patch.get_x() + patch.get_width()
            y0[i] = patch.get_y()
            y1[i] = patch.get_y() + patch.get_height()
        return x0, x1, y0, y1

    def _plt_set_data(self, x0, x1, y0, y1):
        for _ in range(x0.size - len(self.patches)):
            rect = _rectangle(0, 0, 1, 1)
            self.patches.append(rect)
            rect.set_visible(self._visible)
        for _ in range(len(self.patches) - x0.size):
            self.patches.pop()
        for patch, x0i, x1i, y0i, y1i in zip(self.patches, x0, x1, y0, y1):
            patch.set_x(x0i)
            patch.set_width(x1i - x0i)
            patch.set_y(y0i)
            patch.set_height(y1i - y0i)

    ##### HasFace protocol #####

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        if len(self.patches) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([patch.get_facecolor() for patch in self.patches], axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, size=len(self.patches))
        for patch, c in zip(self.patches, color):
            patch.set_facecolor(c)

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [Hatch(patch.get_hatch() or "") for patch in self.patches]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if isinstance(pattern, Hatch):
            pattern = [pattern] * len(self.patches)
        for pat, patch in zip(pattern, self.patches):
            patch.set_hatch(None if pat is Hatch.SOLID else pat.value)

    ##### HasEdges protocol #####

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        if len(self.patches) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([patch.get_edgecolor() for patch in self.patches], axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, size=len(self.patches))
        for patch, c in zip(self.patches, color):
            patch.set_edgecolor(c)

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [LineStyle(patch.get_linestyle()) for patch in self.patches]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            style = [style] * len(self.patches)
        for patch, s in zip(self.patches, style):
            patch.set_linestyle(s.value)

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return np.array([patch.get_linewidth() for patch in self.patches])

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_not_array(width):
            _width = [width] * len(self.patches)
        else:
            _width = width
        for patch, w in zip(self.patches, _width):
            patch.set_linewidth(w)

    def contains(self, event):
        ind: list[int] = []
        for i, patch in enumerate(self.patches):
            contains, _ = patch.contains(event)
            if contains:
                ind.append(i)
        return len(ind) > 0, {"ind": ind}



from __future__ import annotations

import sys
import warnings
import weakref
from typing import TYPE_CHECKING, Callable

import numpy as np
from plotly import graph_objects as go

from whitecanvas import protocols
from whitecanvas.backend.plotly._base import (
    FigureLocation,
    PlotlyHoverableLayer,
    PlotlyLayer,
)
from whitecanvas.backend.plotly._labels import Axis, AxisLabel, Ticks, Title
from whitecanvas.backend.plotly._legend import make_sample_item
from whitecanvas.layers._legend import LegendItem, LegendItemCollection
from whitecanvas.types import Location, MouseEvent
from whitecanvas.utils.normalize import rgba_str_color

if TYPE_CHECKING:
    from plotly._subplots import SubplotXY


class Canvas:
    def __init__(
        self,
        fig: go.Figure,
        *,
        row: int = 0,
        col: int = 0,
        secondary_y: bool = False,
        app: str = "default",
    ):
        self._fig = fig
        self._loc = FigureLocation(row + 1, col + 1, secondary_y)
        self._xaxis = Axis(self, axis="xaxis")
        self._yaxis = Axis(self, axis="yaxis")
        self._xticks = Ticks(self, axis="xaxis")
        self._yticks = Ticks(self, axis="yaxis")
        self._title = Title(self)
        self._xlabel = AxisLabel(self, axis="xaxis")
        self._ylabel = AxisLabel(self, axis="yaxis")

    def _subplot_layout(self) -> SubplotXY:
        try:
            layout = self._fig.get_subplot(**self._loc.asdict())
        except Exception:  # manually wrapped backend are not created with subplots
            layout = self._fig.layout
        return layout

    def _plt_get_native(self):
        return self._fig

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_reorder_layers(self, layers: list[PlotlyLayer]):
        model_to_idx_map = {id(layer._props): i for i, layer in enumerate(layers)}
        data = self._fig._data
        ordered_data = []
        data_in_other = []
        for _data in data:
            data_id = id(_data)
            if data_id in model_to_idx_map:
                ordered_data.append(data[model_to_idx_map[data_id]])
            else:
                data_in_other.append(_data)
        self._fig._data = [*ordered_data, *data_in_other]

    def _plt_get_aspect_ratio(self) -> float | None:
        """Get aspect ratio of canvas"""
        try:
            locked = self._fig["layout"][self._yaxis.name]["scaleanchor"] == "x"
        except KeyError:
            locked = False
        if locked:
            return 1
        return None

    def _plt_set_aspect_ratio(self, ratio: float | None):
        """Set aspect ratio of canvas"""
        if ratio is None:
            self._fig["layout"][self._yaxis.name]["scaleanchor"] = None
        elif ratio == 1:
            self._fig["layout"][self._yaxis.name]["scaleanchor"] = "x"
        else:
            raise NotImplementedError(
                f"Invalid aspect ratio for plotly backend: {ratio}"
            )

    def _plt_add_layer(self, layer: PlotlyLayer):
        self._fig.add_trace(layer._props, **self._loc.asdict())
        layer._props = self._fig.data[-1]
        layer._props["uid"] = layer._props.uid
        if isinstance(layer, PlotlyHoverableLayer):
            layer._connect_mouse_events(self._fig)

    def _plt_remove_layer(self, layer: PlotlyLayer):
        """Remove layer from the canvas"""
        # self._fig._data.remove(layer._props)
        for i, trace in enumerate(self._fig.data):
            if trace["uid"] == layer._props.uid:
                self._fig.data = [*self._fig.data[:i], *self._fig.data[i + 1 :]]
                break
        else:
            raise ValueError(f"Layer {layer._props} not found")

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return True

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        if not visible:
            warnings.warn(
                "Plotly backend does not support hiding canvas",
                UserWarning,
                stacklevel=2,
            )

    def _plt_connect_xlim_changed(self, callback):
        propname = f"{self._xaxis.name}.range"
        self._fig.layout.on_change(lambda _, lim: callback(lim), propname)

    def _plt_connect_ylim_changed(self, callback):
        propname = f"{self._yaxis.name}.range"
        self._fig.layout.on_change(lambda _, lim: callback(lim), propname)

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""
        # TODO

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to drag event"""
        # TODO

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to double-clicked event"""
        # TODO

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

    def _plt_draw(self):
        pass

    def _plt_get_mouse_enabled(self):
        return self._fig.layout.xaxis.fixedrange

    def _plt_set_mouse_enabled(self, enabled: bool):
        self._fig.layout.xaxis.fixedrange = not enabled
        self._fig.layout.yaxis.fixedrange = not enabled

    def _plt_twinx(self):
        from plotly._subplots import SubplotRef

        # count axis numbers
        y_id = 1
        for _c in self._fig._grid_ref:
            for _r in _c:
                for _ in _r:
                    y_id += 1
        if y_id == 1:
            raise RuntimeError("did not find any existing axis.")
        cur_ref = self._fig._grid_ref[self._loc.row - 1][self._loc.col - 1]
        x_id: str = cur_ref[0].layout_keys[0].replace("xaxis", "")
        cur_ref = (
            *cur_ref,
            SubplotRef(
                subplot_type="xy",
                layout_keys=(f"xaxis{x_id}", f"yaxis{y_id}"),
                trace_kwargs={"xaxis": f"x{x_id}", "yaxis": f"y{y_id}"},
            ),
        )
        self._fig._grid_ref[self._loc.row - 1][self._loc.col - 1] = cur_ref
        self._fig.update_layout(
            yaxis2={
                "overlaying": "y",
                "side": "right",
            }
        )
        kwargs = self._loc.asdict()
        kwargs["secondary_y"] = True
        return Canvas(self._fig, **kwargs)

    def _plt_make_legend(
        self,
        items: list[tuple[str, LegendItem]],
        anchor: Location = Location.TOP_RIGHT,
    ):
        plotly_traces = []
        for label, item in items:
            if item is None:
                continue
            if isinstance(item, LegendItemCollection):
                for sub_label, sub_item in item.items:
                    sample = make_sample_item(sub_item)
                    if sample is not None:
                        sample.name = sub_label
                        plotly_traces.append(sample)
            else:
                sample = make_sample_item(item)
                if sample is not None:
                    sample.name = label
                    plotly_traces.append(sample)
        legend_kwargs = _LEGEND_KWARGS[anchor]
        self._fig.add_traces(plotly_traces, **self._loc.asdictn())
        self._fig.update_layout(showlegend=True, legend=legend_kwargs, overwrite=True)

    def _repr_mimebundle_(self, *args, **kwargs):
        return self._fig._repr_mimebundle_(*args, **kwargs)


_LEGEND_KWARGS = {
    Location.TOP_LEFT: {
        "x": 0.01,
        "y": 0.99,
        "xanchor": "left",
        "yanchor": "top",
    },
    Location.TOP_CENTER: {
        "x": 0.5,
        "y": 0.99,
        "xanchor": "center",
        "yanchor": "top",
    },
    Location.TOP_RIGHT: {
        "x": 0.99,
        "y": 0.99,
        "xanchor": "right",
        "yanchor": "top",
    },
    Location.BOTTOM_LEFT: {
        "x": 0.01,
        "y": 0.01,
        "xanchor": "left",
        "yanchor": "bottom",
    },
    Location.BOTTOM_CENTER: {
        "x": 0.5,
        "y": 0.01,
        "xanchor": "center",
        "yanchor": "bottom",
    },
    Location.BOTTOM_RIGHT: {
        "x": 0.99,
        "y": 0.01,
        "xanchor": "right",
        "yanchor": "bottom",
    },
    Location.CENTER_LEFT: {
        "x": 0.01,
        "y": 0.5,
        "xanchor": "left",
        "yanchor": "middle",
    },
    Location.CENTER_RIGHT: {
        "x": 0.99,
        "y": 0.5,
        "xanchor": "right",
        "yanchor": "middle",
    },
    Location.CENTER: {
        "x": 0.5,
        "y": 0.5,
        "xanchor": "center",
        "yanchor": "middle",
    },
    Location.TOP_SIDE_LEFT: {
        "x": 0.01,
        "y": 1.03,
        "xanchor": "left",
        "yanchor": "bottom",
    },
    Location.TOP_SIDE_CENTER: {
        "x": 0.5,
        "y": 1.03,
        "xanchor": "center",
        "yanchor": "bottom",
    },
    Location.TOP_SIDE_RIGHT: {
        "x": 0.99,
        "y": 1.03,
        "xanchor": "right",
        "yanchor": "bottom",
    },
    Location.BOTTOM_SIDE_LEFT: {
        "x": 0.01,
        "y": -0.03,
        "xanchor": "left",
        "yanchor": "top",
    },
    Location.BOTTOM_SIDE_CENTER: {
        "x": 0.5,
        "y": -0.03,
        "xanchor": "center",
        "yanchor": "top",
    },
    Location.BOTTOM_SIDE_RIGHT: {
        "x": 0.99,
        "y": -0.03,
        "xanchor": "right",
        "yanchor": "top",
    },
    Location.LEFT_SIDE_TOP: {
        "x": -0.03,
        "y": 0.99,
        "xanchor": "right",
        "yanchor": "top",
    },
    Location.LEFT_SIDE_CENTER: {
        "x": -0.03,
        "y": 0.5,
        "xanchor": "right",
        "yanchor": "middle",
    },
    Location.LEFT_SIDE_BOTTOM: {
        "x": -0.03,
        "y": 0.01,
        "xanchor": "right",
        "yanchor": "bottom",
    },
    Location.RIGHT_SIDE_TOP: {
        "x": 1.03,
        "y": 0.99,
        "xanchor": "left",
        "yanchor": "top",
    },
    Location.RIGHT_SIDE_CENTER: {
        "x": 1.03,
        "y": 0.5,
        "xanchor": "left",
        "yanchor": "middle",
    },
    Location.RIGHT_SIDE_BOTTOM: {
        "x": 1.03,
        "y": 0.01,
        "xanchor": "left",
        "yanchor": "bottom",
    },
}


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[float], widths: list[float], app: str = "default"):
        from plotly.subplots import make_subplots

        if app == "notebook":
            fig_class = go.FigureWidget
        else:
            fig_class = go.Figure
        self._figs = fig_class(
            make_subplots(
                rows=len(heights),
                cols=len(widths),
                row_heights=heights,
                column_widths=widths,
            ),
        )
        self._figs.update_layout(margin={"l": 6, "r": 6, "t": 30, "b": 6})
        self._app = app
        self._heights = heights
        self._widths = widths

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        if rowspan > 1 or colspan > 1:
            raise NotImplementedError("Plotly backend does not support rowspan/colspan")
        return Canvas(self._figs, row=row, col=col, app=self._app)

    def _plt_add_canvas_3d(self, row: int, col: int, rowspan: int, colspan: int):
        from whitecanvas.backend.plotly.components3d import Canvas3D

        return Canvas3D(self._figs, row=row, col=col, app=self._app)

    def _plt_show(self):
        if self._app in ("qt", "wx", "tk"):
            return NotImplemented
        if self._app == "notebook" and "IPython" in sys.modules:
            from IPython import get_ipython
            from IPython.display import display

            if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
                display(self._figs)
                return
        self._figs.show(renderer="browser")

    def _plt_get_background_color(self):
        return self._figs.layout.paper_bgcolor

    def _plt_set_background_color(self, color):
        self._figs.layout.paper_bgcolor = rgba_str_color(color)

    def _plt_screenshot(self):
        import io

        from PIL import Image

        width, height = self._figs.layout.width, self._figs.layout.height
        img_bytes = self._figs.to_image(format="png", width=width, height=height)
        image = Image.open(io.BytesIO(img_bytes))
        return np.asarray(image, dtype=np.uint8)

    def _plt_set_figsize(self, width: int, height: int):
        self._figs.layout.width = width
        self._figs.layout.height = height

    def _plt_set_spacings(self, wspace: float, hspace: float):
        # plotly does not have a flexible way to set spacings
        pass



__version__ = "0.3.1"

from whitecanvas import theme
from whitecanvas.canvas import link_axes
from whitecanvas.core import (
    new_canvas,
    new_canvas_3d,
    new_col,
    new_grid,
    new_jointgrid,
    new_row,
    wrap_canvas,
)

__all__ = [
    "new_canvas",
    "new_canvas_3d",
    "new_col",
    "new_grid",
    "new_row",
    "new_jointgrid",
    "wrap_canvas",
    "theme",
    "link_axes",
]
